#  2020-27-12 Paul Lettich <paul.lettich@netknights.it>
#             Initial Version of the logging event handler
#
# License:  AGPLv3
# (c) 2016. Cornelius Kölbel
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
__doc__ = """This is the event handler module for logging.
It can be bound to each event and can perform the action:

  * logging: Send a log message to the configured python logging facility

The module is tested in tests/test_lib_eventhandler_logging.py
"""
from privacyidea.lib.eventhandler.base import BaseEventHandler
from privacyidea.lib.token import get_tokens
from privacyidea.lib.utils import create_tag_dict, to_unicode
from privacyidea.lib import _
import logging

log = logging.getLogger(__name__)


class LOGGING_LEVEL(object):
    """
    Allowed logging levels
    """
    ERROR = logging.getLevelName(logging.ERROR)
    WARNING = logging.getLevelName(logging.WARNING)
    INFO = logging.getLevelName(logging.INFO)
    DEBUG = logging.getLevelName(logging.DEBUG)


DEFAULT_LOGMSG = "event={action} triggered"
DEFAULT_LOGGER_NAME = 'pi-eventlogger'
DEFAULT_LOGLEVEL = LOGGING_LEVEL.INFO


class LoggingEventHandler(BaseEventHandler):
    """
    The LoggingEventHandler issues a log message every time the event occurs.
    """

    identifier = "Logging"
    description = "This eventhandler logs the event to the specified python " \
                  "logging facility"

    @property
    def allowed_positions(self):
        """
        This returns the allowed positions of the event handler definition.
        :return: list of allowed positions
        :rtype: list
        """
        return ["post", "pre"]

    @property
    def actions(self):
        """
        This method returns a dictionary of allowed actions and possible
        options in this handler module.

        :return: Dictionary with actions
        :rtype: dict
        """
        actions = {
            "logging": {
                'name': {
                    'type': 'str',
                    'required': False,
                    'default': DEFAULT_LOGGER_NAME,
                    'description': _('The name of the logging facility')
                },
                'message': {
                    'type': 'str',
                    'required': False,
                    'default': DEFAULT_LOGMSG,
                    'description': _('The string to write to the log')
                },
                'level': {
                    'type': 'str',
                    'required': False,
                    'default': DEFAULT_LOGLEVEL,
                    'description': _('The logging level for this logging notification'),
                    'value': [
                        LOGGING_LEVEL.ERROR,
                        LOGGING_LEVEL.WARNING,
                        LOGGING_LEVEL.INFO,
                        LOGGING_LEVEL.DEBUG
                    ]
                }
            }
        }
        return actions

    def do(self, action, options=None):
        """
        This method executes the defined action in the given event.

        :param action:
        :param options: Contains the flask parameters g, request, response
            and the handler_def configuration
        :type options: dict
        :return:
        :rtype: bool
        """
        ret = False

        if action.lower() == 'logging':
            tokentype = None
            g = options.get("g")
            handler_def = options.get("handler_def")
            handler_options = handler_def.get("options", {})
            logged_in_user = g.logged_in_user if hasattr(g, 'logged_in_user') else {}
            request = options.get("request")
            response = options.get("response")
            tokenowner = self._get_tokenowner(request)
            content = self._get_response_content(response)
            serial = (request.all_data.get("serial")
                      or content.get("detail", {}).get("serial")
                      or g.audit_object.audit_data.get("serial"))
            if serial:
                tokens = get_tokens(serial=serial)
                if tokens:
                    tokentype = tokens[0].get_tokentype()
            else:
                token_objects = get_tokens(user=tokenowner)
                serial = ','.join([tok.get_serial() for tok in token_objects])

            tags = create_tag_dict(logged_in_user=logged_in_user,
                                   request=request,
                                   client_ip=g.client_ip,
                                   tokenowner=tokenowner,
                                   serial=serial,
                                   tokentype=tokentype)

            logger_name = handler_options.get('name', DEFAULT_LOGGER_NAME)
            log_action = logging.getLogger(logger_name)
            log_level = getattr(logging,
                                handler_options.get('level', DEFAULT_LOGLEVEL),
                                logging.INFO)
            log_template = handler_options.get('message') or DEFAULT_LOGMSG
            log_action.log(log_level, to_unicode(log_template).format(**tags))
            ret = True

        return ret



"""
This test file tests the lib.tokens.smstoken
"""
import logging

from .base import MyTestCase, FakeFlaskG, FakeAudit
from privacyidea.lib.resolver import (save_resolver)
from privacyidea.lib.realm import (set_realm)
from privacyidea.lib.user import (User)
from privacyidea.lib.utils import is_true
from privacyidea.lib.token import init_token, remove_token
from privacyidea.lib.tokens.smstoken import SmsTokenClass, SMSACTION
from privacyidea.models import (Token, Config, Challenge)
from privacyidea.lib.config import (set_privacyidea_config, set_prepend_pin)
from privacyidea.lib.policy import set_policy, SCOPE, PolicyClass
from privacyidea.lib import _
import datetime
import mock
import responses
from testfixtures import log_capture

PWFILE = "tests/testdata/passwords"


class SMSTokenTestCase(MyTestCase):
    """
    Test the token on the database level
    """
    phone1 = "+49 123456789"
    otppin = "topsecret"
    resolvername1 = "resolver1"
    resolvername2 = "Resolver2"
    resolvername3 = "reso3"
    realm1 = "realm1"
    realm2 = "realm2"
    serial1 = "SE123456"
    serial2 = "SE222222"
    otpkey = "3132333435363738393031323334353637383930"

    SMSHttpUrl = "http://smsgateway.com/sms_send_api.cgi"
    SMSProviderConfig = '''{"URL": "http://smsgateway.com/sms_send_api.cgi",
                   "PARAMETER": {"from": "0170111111",
                                 "password": "yoursecret",
                                 "sender": "name",
                                 "account": "company_ltd"},
                   "SMS_TEXT_KEY": "text",
                   "SMS_PHONENUMBER_KEY": "destination",
                   "HTTP_Method": "POST",
                   "PROXY": "http://username:password@your-proxy:8080",
                   "RETURN_SUCCESS": "ID"
    }'''
    success_body = "ID 12345"


    # add_user, get_user, reset, set_user_identifiers

    def test_00_create_user_realm(self):
        rid = save_resolver({"resolver": self.resolvername1,
                               "type": "passwdresolver",
                               "fileName": PWFILE})
        self.assertTrue(rid > 0, rid)

        (added, failed) = set_realm(self.realm1, [{'name': self.resolvername1}])
        self.assertTrue(len(failed) == 0)
        self.assertTrue(len(added) == 1)

        user = User(login="root",
                    realm=self.realm1,
                    resolver=self.resolvername1)

        user_str = "{0!s}".format(user)
        self.assertTrue(user_str == "<root.resolver1@realm1>", user_str)

        self.assertFalse(user.is_empty())
        self.assertTrue(User().is_empty())

        user_repr = "{0!r}".format(user)
        expected = "User(login='root', realm='realm1', resolver='resolver1')"
        self.assertTrue(user_repr == expected, user_repr)

    def test_01_create_token(self):
        db_token = Token(self.serial1, tokentype="sms")
        db_token.save()
        token = SmsTokenClass(db_token)
        token.update({"phone": self.phone1})
        token.save()
        self.assertTrue(token.token.serial == self.serial1, token)
        self.assertTrue(token.token.tokentype == "sms", token.token)
        self.assertTrue(token.type == "sms", token.type)
        class_prefix = token.get_class_prefix()
        self.assertTrue(class_prefix == "PISM", class_prefix)
        self.assertTrue(token.get_class_type() == "sms", token)

        db_token = Token(self.serial2, tokentype="sms")
        db_token.save()
        token = SmsTokenClass(db_token)
        token.update({"dynamic_phone": True})
        token.save()
        self.assertTrue(token.token.serial == self.serial2, token)
        self.assertTrue(token.token.tokentype == "sms", token.token)
        self.assertTrue(is_true(token.get_tokeninfo("dynamic_phone")))
        self.assertTrue(token.type == "sms", token.type)
        class_prefix = token.get_class_prefix()
        self.assertTrue(class_prefix == "PISM", class_prefix)
        self.assertTrue(token.get_class_type() == "sms", token)
        token.add_user(User(login="cornelius",
                            realm=self.realm1))

    def test_02_set_user(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = SmsTokenClass(db_token)
        self.assertTrue(token.token.tokentype == "sms",
                        token.token.tokentype)
        self.assertTrue(token.type == "sms", token.type)

        token.add_user(User(login="cornelius",
                            realm=self.realm1))
        self.assertEqual(token.token.first_owner.resolver, self.resolvername1)
        self.assertEqual(token.token.first_owner.user_id, "1000")

        user_object = token.user
        self.assertTrue(user_object.login == "cornelius",
                        user_object)
        self.assertTrue(user_object.resolver == self.resolvername1,
                        user_object)

    def test_03_reset_failcounter(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = SmsTokenClass(db_token)
        token.token.failcount = 10
        token.reset()
        self.assertTrue(token.token.failcount == 0,
                        token.token.failcount)

    def test_04_base_methods(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = SmsTokenClass(db_token)
        self.assertTrue(token.check_otp("123456", 1, 10) == -1)

        # get class info
        cli = token.get_class_info()
        self.assertTrue(cli.get("type") == "sms", cli.get("type"))
        cli = token.get_class_info("type")
        self.assertTrue(cli == "sms", cli)

        # set the description
        token.set_description("something new")
        self.assertTrue(token.token.description == "something new",
                        token.token)

        # set defaults
        token.set_defaults()
        self.assertTrue(token.token.otplen == 6)
        self.assertTrue(token.token.sync_window == 1000)

        token.resync("1234", "3456")

        token.token.count_window = 17
        self.assertTrue(token.get_otp_count_window() == 17)

        token.token.count = 18
        self.assertTrue(token.get_otp_count() == 18)

        token.token.active = False
        self.assertTrue(token.is_active() is False)

        token.token.failcount = 7
        self.assertTrue(token.get_failcount() == 7)
        token.set_failcount(8)
        self.assertTrue(token.token.failcount == 8)

        token.token.maxfail = 12
        self.assertTrue(token.get_max_failcount() == 12)

        self.assertEqual(token.get_user_id(), token.token.first_owner.user_id)

        self.assertTrue(token.get_serial() == "SE123456", token.token.serial)
        self.assertTrue(token.get_tokentype() == "sms",
                        token.token.tokentype)

        token.set_so_pin("sopin")
        token.set_user_pin("userpin")
        token.set_otpkey(self.otpkey)
        token.set_otplen(8)
        token.set_otp_count(1000)
        self.assertTrue(len(token.token.so_pin) == 32,
                        token.token.so_pin)
        self.assertTrue(len(token.token.user_pin) == 32,
                        token.token.user_pin)
        self.assertEqual(len(token.token.key_enc), 96, token.token.key_enc)
        self.assertTrue(token.get_otplen() == 8)
        self.assertTrue(token.token.count == 1000,
                        token.token.count)

        token.set_maxfail(1000)
        self.assertTrue(token.token.maxfail == 1000)

        token.set_count_window(52)
        self.assertTrue(token.get_count_window() == 52)

        token.set_sync_window(53)
        self.assertTrue(token.get_sync_window() == 53)

    def test_05_get_set_realms(self):
        set_realm(self.realm2)
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = SmsTokenClass(db_token)
        realms = token.get_realms()
        self.assertTrue(len(realms) == 1, realms)
        token.set_realms([self.realm1, self.realm2])
        realms = token.get_realms()
        self.assertTrue(len(realms) == 2, realms)

    def test_06_set_pin(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = SmsTokenClass(db_token)
        token.set_pin("hallo")
        (ph1, pseed) = token.get_pin_hash_seed()
        # check the database
        token.set_pin("blubber")
        ph2 = token.token.pin_hash
        self.assertTrue(ph1 != ph2)
        token.set_pin_hash_seed(ph1, pseed)

    def test_07_enable(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = SmsTokenClass(db_token)
        token.enable(False)
        self.assertFalse(token.token.active)
        token.enable()
        self.assertTrue(token.token.active)
        # we need to safe the token explicitly here to persist it to the db
        token.save()

    def test_08_info(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = SmsTokenClass(db_token)
        token.set_hashlib("sha1")
        ti = token.get_tokeninfo()
        self.assertTrue("hashlib" in ti, ti)

    def test_09_failcount(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = SmsTokenClass(db_token)
        start = token.token.failcount
        end = token.inc_failcount()
        self.assertTrue(end == start + 1, (end, start))

    def test_10_get_hashlib(self):
        # check if functions are returned
        for hl in ["sha1", "md5", "sha256", "sha512",
                   "sha224", "sha384", "", None]:
            self.assertTrue(hasattr(SmsTokenClass.get_hashlib(hl),
                                    '__call__'),
                            SmsTokenClass.get_hashlib(hl))

    def test_11_tokeninfo(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = SmsTokenClass(db_token)
        token.add_tokeninfo("key1", "value2")
        info1 = token.get_tokeninfo()
        self.assertTrue("key1" in info1, info1)
        token.add_tokeninfo("key2", "value3")
        info2 = token.get_tokeninfo()
        self.assertTrue("key2" in info2, info2)
        token.set_tokeninfo(info1)
        info2 = token.get_tokeninfo()
        self.assertTrue("key2" not in info2, info2)
        self.assertTrue(token.get_tokeninfo("key1") == "value2",
                        info2)

        # auth counter
        token.set_count_auth_success_max(200)
        token.set_count_auth_max(1000)
        token.set_count_auth_success(100)
        token.inc_count_auth_success()
        token.set_count_auth(200)
        token.inc_count_auth()
        self.assertTrue(token.get_count_auth_success_max() == 200)
        self.assertTrue(token.get_count_auth_success() == 101)
        self.assertTrue(token.get_count_auth_max() == 1000)
        self.assertTrue(token.get_count_auth() == 201)

        self.assertTrue(token.check_auth_counter())
        token.set_count_auth_max(10)
        self.assertFalse(token.check_auth_counter())
        token.set_count_auth_max(1000)
        token.set_count_auth_success_max(10)
        self.assertFalse(token.check_auth_counter())

    def test_12_inc_otp_counter(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = SmsTokenClass(db_token)

        token.set_otp_count(10)
        self.assertTrue(token.token.count == 10, token.token.count)
        # increase counter by 1
        token.inc_otp_counter()
        self.assertTrue(token.token.count == 11, token.token.count)
        # increase counter to 21
        Config(Key="DefaultResetFailCount", Value=True).save()
        token.inc_otp_counter(counter=20)
        self.assertTrue(token.token.count == 21, token.token.count)

    def test_13_check_otp(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = SmsTokenClass(db_token)
        token.update({"otpkey": self.otpkey,
                      "pin": "test",
                      "otplen": 6,
                      "phone": self.phone1})
        # OTP does not exist
        self.assertTrue(token.check_otp_exist("222333") == -1)
        # OTP does exist
        res = token.check_otp_exist("969429")
        self.assertTrue(res == 3, res)

    def test_14_split_pin_pass(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = SmsTokenClass(db_token)

        token.token.otplen = 6
        # postpend pin
        set_prepend_pin(False)
        _res, pin, value = token.split_pin_pass("222333test")
        self.assertTrue(pin == "test", pin)
        self.assertTrue(value == "222333", value)
        # prepend pin
        set_prepend_pin(True)
        _res, pin, value = token.split_pin_pass("test222333")
        self.assertTrue(pin == "test", pin)
        self.assertTrue(value == "222333", value)

    def test_15_check_pin(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = SmsTokenClass(db_token)
        # test the encrypted pin
        token.set_pin("encrypted", encrypt=True)
        self.assertTrue(token.check_pin("encrypted"))
        self.assertFalse(token.check_pin("wrong pin"))

        # test the hashed pin
        token.set_pin("test")
        self.assertTrue(token.check_pin("test"))
        self.assertFalse(token.check_pin("wrong pin"))

    def test_17_challenge_token(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = SmsTokenClass(db_token)
        token.set_pin(self.otppin)

        r = token.is_challenge_request(self.otppin)
        self.assertTrue(r)

    @responses.activate
    def test_18_challenge_request(self):
        responses.add(responses.POST,
                      self.SMSHttpUrl,
                      body=self.success_body)
        transactionid = "123456098712"
        set_privacyidea_config("sms.providerConfig", self.SMSProviderConfig)
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = SmsTokenClass(db_token)
        self.assertTrue(token.check_otp("123456", 1, 10) == -1)
        c = token.create_challenge(transactionid)
        self.assertTrue(c[0], c)
        otp = c[1]
        self.assertTrue(c[3]["attributes"]["state"], transactionid)

        # check for the challenges response
        r = token.check_challenge_response(passw=otp,
                                           options={"transaction_id": transactionid})
        self.assertTrue(r, r)

    @responses.activate
    def test_18a_challenge_request_dynamic(self):
        # Send a challenge request for an SMS token with a dynamic phone number
        responses.add(responses.POST,
                      self.SMSHttpUrl,
                      body=self.success_body)
        transactionid = "123456098712"
        set_privacyidea_config("sms.providerConfig", self.SMSProviderConfig)
        db_token = Token.query.filter_by(serial=self.serial2).first()
        token = SmsTokenClass(db_token)
        self.assertTrue(token.check_otp("123456", 1, 10) == -1)
        c = token.create_challenge(transactionid)
        self.assertTrue(c[0], c)
        otp = c[1]
        self.assertTrue(c[3]["attributes"]["state"], transactionid)

        # check for the challenges response
        r = token.check_challenge_response(passw=otp,
                                           options={"transaction_id": transactionid})
        self.assertTrue(r, r)

    @responses.activate
    def test_18b_challenge_request_dynamic_multivalue(self):
        responses.add(responses.POST,
                      self.SMSHttpUrl,
                      body=self.success_body)
        transactionid = "123456098712"
        set_privacyidea_config("sms.providerConfig", self.SMSProviderConfig)
        db_token = Token.query.filter_by(serial=self.serial2).first()
        token = SmsTokenClass(db_token)
        # if the email is a multi-value attribute, the first address should be chosen
        new_user_info = token.user.info.copy()
        new_user_info['mobile'] = ['1234', '5678']
        with mock.patch('privacyidea.lib.resolvers.PasswdIdResolver.IdResolver.getUserInfo') as mock_user_info:
            mock_user_info.return_value = new_user_info
            c = token.create_challenge(transactionid)
            self.assertTrue(c[0], c)
            self.assertIn('destination=1234', responses.calls[0].request.body)
            self.assertNotIn('destination=5678', responses.calls[0].request.body)

    @responses.activate
    def test_19_smstext(self):
        # The single quotes in the smstext "'Your <otp>'" is legacy and results in
        # the string without single quotes "Your <otp>".
        smstext_tests = {"'Your <otp>'": r"Your [0-9]{6}",
                         "Your <otp>": r"Your [0-9]{6}",
                         "{user} has the OTP: {otp}": r"Cornelius has the OTP: [0-9]{6}"}
        for pol_text, result_text in smstext_tests.items():
            # create a SMSTEXT policy:
            p = set_policy(name="smstext",
                           action="{0!s}={1!s}".format(SMSACTION.SMSTEXT, pol_text),
                           scope=SCOPE.AUTH)
            self.assertTrue(p > 0)

            g = FakeFlaskG()
            P = PolicyClass()
            g.audit_object = FakeAudit()
            g.policy_object = P
            options = {"g": g,
                       "user": User("cornelius", self.realm1)}

            responses.add(responses.POST,
                          self.SMSHttpUrl,
                          body=self.success_body)
            set_privacyidea_config("sms.providerConfig", self.SMSProviderConfig)
            db_token = Token.query.filter_by(serial=self.serial1).first()
            token = SmsTokenClass(db_token)
            c = token.create_challenge(options=options)
            self.assertTrue(c[0], c)
            display_message = c[1]
            self.assertEqual(display_message, _("Enter the OTP from the SMS:"))
            self.assertEqual(c[3].get("state"), None)

            smstext = token._get_sms_text(options)
            self.assertEqual(pol_text.strip("'"), smstext)
            r, message = token._send_sms(smstext, options)
            self.assertRegex(message, result_text)

        # Test AUTOSMS
        p = set_policy(name="autosms",
                       action=SMSACTION.SMSAUTO,
                       scope=SCOPE.AUTH)
        self.assertTrue(p > 0)

        g = FakeFlaskG()
        P = PolicyClass()
        g.policy_object = P
        g.audit_object = FakeAudit()
        options = {"g": g}

        r = token.check_otp(self.valid_otp_values[5 + len(smstext_tests)], options=options)
        self.assertTrue(r > 0, r)

    @ log_capture(level=logging.WARN)
    def test_21_failed_loading(self, capture):
        token = init_token({'type': 'sms', 'phone': self.phone1})
        transactionid = "123456098712"
        set_privacyidea_config("sms.providerConfig", "noJSON")
        set_privacyidea_config("sms.provider",
                               "privacyidea.lib.smsprovider."
                               "HttpSMSProvider.HttpSMSProviderWRONG")

        with mock.patch("logging.Logger.error") as mock_log:
            c = token.create_challenge(transactionid)
            self.assertFalse(c[0], c)
            self.assertTrue(c[1].startswith("The PIN was correct, but"), c[1])
            expected = "Failed to load SMSProvider: ImportError" \
                       "('privacyidea.lib.smsprovider.HttpSMSProvider has no attribute HttpSMSProviderWRONG'"
            mock_log.mock_called()
            mocked_str = mock_log
            self.assertTrue(mocked_str.startswith(expected), mocked_str)
        capture.clear()

        with mock.patch("logging.Logger.error") as mock_log:
            set_privacyidea_config("sms.provider",
                                   "privacyidea.lib.smsprovider."
                                   "HttpSMSProvider.HttpSMSProvider")
            c = token.create_challenge(transactionid)
            self.assertFalse(c[0], c)
            self.assertTrue(c[1].startswith("The PIN was correct, but"), c[1])
            expected = "Failed to load sms.providerConfig: " \
                       "JSONDecodeError('Expecting value: line 1 column 1 (char 0)')"
            mock_log.mock_called()
            mocked_str = mock_log
            self.assertTrue(mocked_str.startswith(expected), mocked_str)
        capture.clear()

        # test with the parameter exception=1
        self.assertRaises(Exception, token.create_challenge, transactionid, {"exception": "1"})

        remove_token(token.get_serial())

    def test_99_delete_token(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = SmsTokenClass(db_token)
        token.delete_token()

        db_token = Token.query.filter_by(serial=self.serial1).first()
        self.assertTrue(db_token is None, db_token)



#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  2018-01-15   Friedrich Weber <friedrich.weber@netknights.it>
#               Initial version of the VASCO token
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
import binascii

__doc__ = """This is the implementation of the VASCO token"""

import logging
from privacyidea.api.lib.utils import getParam
from privacyidea.lib.utils import is_true
from privacyidea.lib.decorators import check_token_locked
from privacyidea.lib.error import ParameterError
from privacyidea.lib.log import log_with
from privacyidea.lib.tokenclass import TokenClass
from privacyidea.lib.tokens.vasco import vasco_otp_check
from privacyidea.lib import _
from privacyidea.lib.policy import SCOPE, ACTION, GROUP

optional = True
required = False

log = logging.getLogger(__name__)


class VascoTokenClass(TokenClass):
    """
    Token class for VASCO Digipass tokens. Relies on vendor-specific
    shared library, whose location needs to be set in the PI_VASCO_LIBRARY
    config option.

    VASCO Tokens can be read from a CSV file which is structured as follows::

        <serial1>,<hexlify(blob1)>,vasco
        <serial2>,<hexlify(blob2)>,vasco
        ...

    whereas blobX is the 248-byte blob holding the token information.
    Consequently, hexlify(blobX) is a 496-character hex string.

    The CSV file can be imported by using the "Import Tokens" feature of the Web UI,
    where "OATH CSV" needs to be chosen as the file type.
    """

    def __init__(self, db_token):
        """
        constructor - create a token class object with its db token binding

        :param aToken: the db bound token
        """
        TokenClass.__init__(self, db_token)
        self.set_type("vasco")
        self.hKeyRequired = True

    @staticmethod
    def get_class_type():
        """
        return the class type identifier
        """
        return "vasco"

    @staticmethod
    def get_class_prefix():
        """
        return the token type prefix
        """
        # TODO: Revisit token type?
        return "VASC"

    @staticmethod
    @log_with(log)
    def get_class_info(key=None, ret='all'):
        """
        :param key: subsection identifier
        :type key: string
        :param ret: default return value, if nothing is found
        :type ret: user defined
        :return: subsection if key exists or user defined
        :rtype: dict or string
        """
        res = {'type': 'vasco',
               'title': 'VASCO Token',
               'description': _('VASCO Token: Authentication using VASCO tokens'),
               # If this was set, the user could enroll a Vasco token via the API
               #'user': ["enroll"],
               # only administrators can enroll the token in the UI
               'ui_enroll': ["admin"],
               'policy': {
                   SCOPE.ENROLL: {
                       ACTION.MAXTOKENUSER: {
                           'type': 'int',
                           'desc': _("The user may only have this maximum number of Vasco tokens assigned."),
                           'group': GROUP.TOKEN
                       },
                       ACTION.MAXACTIVETOKENUSER: {
                           'type': 'int',
                           'desc': _(
                               "The user may only have this maximum number of active Vasco tokens assigned."),
                           'group': GROUP.TOKEN
                       }
                   }
               },
               }

        if key:
            ret = res.get(key, {})
        else:
            if ret == 'all':
                ret = res
        return ret

    @log_with(log)
    def update(self, param, reset_failcount=True):
        """
        update - process initialization parameters

        :param param: dict of initialization parameters
        :type param: dict

        :return: nothing
        """
        if is_true(getParam(param, 'genkey', optional)):
            raise ParameterError("Generating OTP keys is not supported")

        upd_param = param.copy()

        # If the OTP key is given, it is given as a 496-character hex string which
        # encodes a 248-byte blob. As we want to set a 248-byte OTPKey (= Blob),
        # we unhexlify the OTP key
        if 'otpkey' in param:
            if len(param['otpkey']) != 496:
                raise ParameterError('Expected OTP key as 496-character hex string, but length is {!s}'.format(
                    len(param['otpkey'])
                ))
            try:
                upd_param['otpkey'] = binascii.unhexlify(upd_param['otpkey'])
            except (binascii.Error, TypeError):
                raise ParameterError('Expected OTP key as 496-character hex string, but it is malformed')

        TokenClass.update(self, upd_param, reset_failcount)

    @check_token_locked
    def check_otp(self, otpval, counter=None, window=None, options=None):
        secret = self.token.get_otpkey().getKey()
        # vasco_otp_check expects a bytestring, so we encode ``otpval`` to bytes (should be ASCII anyway)
        result, new_secret = vasco_otp_check(secret, otpval.encode("utf-8"))
        # By default, setting a new OTP key resets the failcounter. In case of the VASCO token,
        # this would mean that the failcounter is reset at every authentication attempt
        # (regardless of success or failure), which must be avoided.
        self.token.set_otpkey(new_secret, reset_failcount=False)
        self.save()

        if result == 0:
            # Successful authentication
            return 0
        else:
            if result == 1:
                # wrong OTP value, no log message
                pass
            elif result == 201:
                log.warning("VASCO token failed to authenticate, code replay attempt, previous OTP value was used again!")
            elif result == 202:
                log.warning("Token-internal fail counter reached its maximum!")
            elif result == -202:
                log.warning("VASCO token failed to authenticate, response too small, user did not type his complete OTP!")
            elif result == -205:
                log.warning("VASCO token failed to authenticate, response not decimal!")
            else:
                log.warning("VASCO token failed to authenticate, result: {!r}".format(result))
            return -1



"""
This test file tests the lib.tokens.papertoken
This depends on lib.tokenclass
"""
import logging

from testfixtures import LogCapture
from .base import MyTestCase
from privacyidea.lib.tokens.tantoken import TanTokenClass
from privacyidea.lib.token import init_token, import_token
from privacyidea.models import Token

OTPKEY = "3132333435363738393031323334353637383930"


class TanTokenTestCase(MyTestCase):
    serial1 = "ser1"

    # add_user, get_user, reset, set_user_identifiers

    def test_01_create_token(self):
        db_token = Token(self.serial1, tokentype="tan")
        db_token.save()
        token = TanTokenClass(db_token)
        token.update({})
        self.assertEqual(token.token.serial, self.serial1)
        self.assertEqual(token.token.tokentype, "tan")
        self.assertEqual(token.type, "tan")
        class_prefix = token.get_class_prefix()
        self.assertEqual(class_prefix, "PITN")
        self.assertEqual(token.get_class_type(), "tan")

    def test_02_class_methods(self):
        db_token = Token.query.filter(Token.serial == self.serial1).first()
        token = TanTokenClass(db_token)

        info = token.get_class_info()
        self.assertEqual(info.get("title"), "TAN Token")

        info = token.get_class_info("title")
        self.assertEqual(info, "TAN Token")

    def test_03_get_init_details(self):
        db_token = Token.query.filter(Token.serial == self.serial1).first()
        token = TanTokenClass(db_token)
        token.update({})

        # if no otpkey was given, an OTP key is created.
        init_detail = token.get_init_detail()
        self.assertTrue("otps" in init_detail)

    def test_04_get_init_details_with_key(self):
        token = init_token({"type": "tan",
                            "otpkey": OTPKEY})
        init_detail = token.get_init_detail()
        self.assertTrue("otps" in init_detail)
        otps = init_detail.get("otps")
        self.assertEqual(otps.get(0), "755224")
        self.assertEqual(otps.get(1), "287082")
        self.assertEqual(otps.get(2), "359152")

        # test authenticating in arbitrary order
        r = token.check_otp("287082")
        self.assertEqual(r, 1)
        r = token.check_otp("359152")
        self.assertEqual(r, 1)
        r = token.check_otp("755224")
        self.assertEqual(r, 1)
        # previous OTP
        r = token.check_otp("755224")
        self.assertEqual(r, -1)
        with LogCapture(level=logging.INFO) as lc:
            self.assertEqual(-1, token.check_otp("12345"))
            lc.check_present(
                ('privacyidea.lib.decorators', 'INFO',
                 f'OTP value for token {token.token.serial} (type: {token.type}) '
                 f'has wrong length (5 != 6)'))
            self.assertEqual(-1, token.check_otp("1234567"))
            lc.check_present(
                ('privacyidea.lib.decorators', 'INFO',
                 f'OTP value for token {token.token.serial} (type: {token.type}) '
                 f'has wrong length (7 != 6)'))

    def test_05_import(self):
        params = TanTokenClass.get_import_csv(["se1", "121212", "tan",
                                               "tan1 tan2  tan3    tan4    tan5"])
        self.assertEqual(params.get("serial"), "se1")
        self.assertEqual(params.get("type"), "tan")
        self.assertEqual(params.get("tans").split(), ["tan1", "tan2", "tan3", "tan4", "tan5"])

        # test init token
        tok = init_token(params)
        self.assertEqual(tok.token.tokentype, "tan")

        d = tok.get_as_dict()
        self.assertTrue("tan.tan0" not in d.get("info"), d.get("info"))
        self.assertEqual(d.get("info", {}).get("tan.count"), 5)

        # check all tans
        r = tok.check_otp("tan2")
        self.assertEqual(r, 1)
        r = tok.check_otp("tan2")
        self.assertEqual(r, -1)
        r = tok.check_otp("tan5")
        self.assertEqual(r, 1)
        r = tok.check_otp("tan4")
        self.assertEqual(r, 1)
        r = tok.check_otp("tan3")
        self.assertEqual(r, 1)

        d = tok.get_as_dict()
        self.assertEqual(d.get("info", {}).get("tan.count"), 1)

        # Check the authentication of a TAN token with a PIN
        tok.set_pin("test")
        r = tok.authenticate("testtan1")
        self.assertEqual(r, (True, 1, None))

        # check if the otplen of the TAN token is 4, the length of the first TAN
        self.assertEqual(tok.token.otplen, 4)

    def test_10_import_tan(self):
        # These are example values that are imported from a file and end up at
        # lib/token.import_token
        serial = "ABC123"
        params = {'hashlib': 'sha1',
                  'type': 'tan',
                  'user': {},
                  'serial': 'ABC123',
                  'tans': '123465 798111',
                  'otpkey': ''}

        tok = import_token(serial, params)
        self.assertEqual("tan", tok.type)

        d = tok.get_as_dict()
        self.assertEqual(d.get("info", {}).get("tan.count"), 2)
        # check tans:
        r = tok.check_otp("798111")
        self.assertEqual(r, 1)
        r = tok.check_otp("798111")
        self.assertEqual(r, -1)
        r = tok.check_otp("123465")
        self.assertEqual(r, 1)



from privacyidea.lib.challenge import get_challenges
from privacyidea.lib.framework import get_app_local_store
from privacyidea.lib.tokenclass import CHALLENGE_SESSION
from .base import MyApiTestCase
from privacyidea.lib.user import (User)
from privacyidea.lib.config import (set_privacyidea_config)
from privacyidea.lib.token import (get_tokens, init_token, remove_token)
from privacyidea.lib.policy import (SCOPE, set_policy, delete_policy)
from privacyidea.lib.smsprovider.SMSProvider import set_smsgateway
from privacyidea.lib.smsprovider.FirebaseProvider import FIREBASE_CONFIG
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from privacyidea.lib.utils import to_bytes, to_unicode
from privacyidea.lib.tokens.pushtoken import (PUSH_ACTION,
                                              strip_key,
                                              PUBLIC_KEY_SMARTPHONE, PRIVATE_KEY_SERVER,
                                              PUBLIC_KEY_SERVER,
                                              POLL_ONLY)
from privacyidea.lib.utils import b32encode_and_unicode
from datetime import datetime
from base64 import b32encode, b64encode, b64decode
from hashlib import sha1
import hmac
import mock
import responses
import urllib
from .test_lib_tokens_push import _check_firebase_params, _create_credential_mock

PWFILE = "tests/testdata/passwords"
HOSTSFILE = "tests/testdata/hosts"
DICT_FILE = "tests/testdata/dictionary"
FIREBASE_FILE = "tests/testdata/firebase-test.json"
CLIENT_FILE = "tests/testdata/google-services.json"
FB_CONFIG_VALS = {
    FIREBASE_CONFIG.JSON_CONFIG: FIREBASE_FILE}
REGISTRATION_URL = "http://test/ttype/push"
TTL = "10"


class TtypeAPITestCase(MyApiTestCase):
    """
    test the api.ttype endpoints
    """

    def test_00_create_realms(self):
        self.setUp_user_realms()

    def test_01_tiqr(self):
        init_token({"serial": "TIQR1",
                    "type": "tiqr"}, User("cornelius", self.realm1))
        with self.app.test_request_context('/ttype/tiqr',
                                           method='POST',
                                           data={"action": "metadata",
                                                 "serial": "TIQR1",
                                                 "session": "12345"}):
            res = self.app.full_dispatch_request()
            data = res.json
            identity = data.get("identity")
            service = data.get("service")
            self.assertEqual(identity.get("displayName"), "Cornelius ")
            self.assertEqual(service.get("displayName"), "privacyIDEA")

    def test_02_u2f(self):
        set_privacyidea_config("u2f.appId", "https://puck.az.intern")
        with self.app.test_request_context('/ttype/u2f',
                                           method='GET'):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            self.assertEqual(res.mimetype, 'application/fido.trusted-apps+json')
            data = res.json
            self.assertTrue("trustedFacets" in data)

        # Check the audit log.
        with self.app.test_request_context('/audit/?action=*GET /ttype/*',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            json_response = res.json
            result = json_response.get("result")
            auditdata = result.get("value").get("auditdata")
            self.assertTrue(len(auditdata) > 0)
            self.assertEqual(auditdata[0].get("token_type"), "u2f")

    def test_03_wrong(self):
        # Test the ttype endpoint for wrong ttype, here /ttype/wrong
        init_token({"serial": "TIQR1",
                    "type": "tiqr"}, User("cornelius", self.realm1))
        with self.app.test_request_context('/ttype/wrong',
                                           method='POST',
                                           data={"action": "metadata",
                                                 "serial": "TIQR1",
                                                 "session": "12345"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 400, res.status_code)

    def test_04_yubikey(self):
        # ---- Verify signature with test vectors from yubico ----
        init_token({"type": "yubikey",
                    "otpkey": "000102030405060708090a0b0c0d0e0f",
                    "otplen": 44,
                    "yubikey.prefix": "vvungrrdhvtklk"})

        # Add API key to the config
        api_1_key = "mG5be6ZJU1qBGz24yPh/ESM3UdU="
        set_privacyidea_config("yubikey.apiid.1", api_1_key)

        # Test the ttype endpoint for yubikey
        parameters = {"id": "1", "otp": "vvungrrdhvtklknvrtvuvbbkeidikkvgglrvdgrfcdft",
                      "nonce": "jrFwbaYFhn0HoxZIsd9LQ6w2ceU",
                      "h": "%2Bja8S3IjbX593%2FLAgTBixwPNGX4%3D"}
        with self.app.test_request_context('/ttype/yubikey',
                                           query_string=parameters,
                                           method='GET'):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            # The signature is valid, but it is not a valid OTP
            self.assertIn(b"status=BAD_OTP", res.data)

        # ---- Verify signature + otp successfully ----
        init_token({"type": "yubikey",
                    "otpkey": "000102030405060708090a0b0c0d0e0f",
                    "otplen": 44,
                    "yubikey.prefix": "vvungrrdhvtk"})

        # Test the ttype endpoint for yubikey
        parameters = {"id": "1", "otp": "vvungrrdhvtkdvgtiblfkbgturecfllberrvkinnctnn",
                      "nonce": "jrFwbaYFhn0HoxZIsd9LQ6w2ceU"}
        # sign parameters
        data_string = f"id={parameters['id']}&nonce={parameters['nonce']}&otp={parameters['otp']}"
        signature = b64encode(to_bytes(hmac.new(b64decode(api_1_key), to_bytes(data_string), sha1).digest()))
        parameters["h"] = urllib.parse.quote_plus(signature.decode("utf-8"))
        with self.app.test_request_context('/ttype/yubikey',
                                           query_string=parameters,
                                           method='GET'):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            # The signature is valid, but it is not a valid OTP
            self.assertIn(b"status=OK", res.data)

        # ---- Check OTP without sending h (signature) ----
        # create yubikey token
        init_token({"type": "yubikey",
                    "otpkey": "cc17a4d77eaed96e9d14b5c87a02e718",
                    "otplen": 48,
                    "yubikey.prefix": "ebedeeefegeheiej"})

        # Add API key to the config
        set_privacyidea_config("yubikey.apiid.2", "1YMEbMZijD3DzL21UfKGnOOI13c=")

        # Test the ttype endpoint for yubikey
        parameters = {"id": "2", "otp": "ebedeeefegeheiejtjtrutblehenfjljrirgdihrfuetljtt",
                      "nonce": "random nonce"}
        with self.app.test_request_context('/ttype/yubikey',
                                           query_string=parameters,
                                           method='GET'):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            self.assertIn(b"status=OK", res.data)


class TtypePushAPITestCase(MyApiTestCase):
    """
    test /ttype/push
    """

    server_private_key = rsa.generate_private_key(public_exponent=65537,
                                                  key_size=4096,
                                                  backend=default_backend())
    server_private_key_pem = to_unicode(server_private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption()))
    server_public_key_pem = to_unicode(server_private_key.public_key().public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo))

    # We now allow white spaces in the firebase config name
    firebase_config_name = "my firebase config"

    smartphone_private_key = rsa.generate_private_key(public_exponent=65537,
                                                      key_size=4096,
                                                      backend=default_backend())
    smartphone_public_key = smartphone_private_key.public_key()
    smartphone_public_key_pem = to_unicode(smartphone_public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo))
    # The smartphone sends the public key in URLsafe and without the ----BEGIN header
    smartphone_public_key_pem_urlsafe = strip_key(smartphone_public_key_pem).replace("+", "-").replace("/", "_")
    serial_push = "PIPU001"

    def _create_push_token(self):
        tparams = {'type': 'push', 'genkey': 1}
        tparams.update(FB_CONFIG_VALS)
        tok = init_token(param=tparams)
        tok.add_tokeninfo(PUSH_ACTION.FIREBASE_CONFIG, self.firebase_config_name)
        tok.add_tokeninfo(PUBLIC_KEY_SMARTPHONE, self.smartphone_public_key_pem_urlsafe)
        tok.add_tokeninfo('firebase_token', 'firebaseT')
        tok.add_tokeninfo(PUBLIC_KEY_SERVER, self.server_public_key_pem)
        tok.add_tokeninfo(PRIVATE_KEY_SERVER, self.server_private_key_pem, 'password')
        tok.del_tokeninfo("enrollment_credential")
        tok.token.rollout_state = "enrolled"
        tok.token.active = True
        return tok

    def test_00_create_realms(self):
        self.setUp_user_realms()

    def test_01_api_enroll_push(self):
        self.authenticate()

        # Failed enrollment due to missing policy
        with self.app.test_request_context('/token/init',
                                           method='POST',
                                           data={"type": "push",
                                                 "genkey": 1},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertNotEqual(res.status_code, 200)
            error = res.json.get("result").get("error")
            self.assertEqual(error.get("message"),
                             "Missing enrollment policy for push token: push_firebase_configuration")
            self.assertEqual(error.get("code"), 303)

        r = set_smsgateway(self.firebase_config_name,
                           'privacyidea.lib.smsprovider.FirebaseProvider.FirebaseProvider',
                           "myFB", FB_CONFIG_VALS)
        self.assertTrue(r > 0)
        set_policy("push1", scope=SCOPE.ENROLL,
                   action="{0!s}={1!s},{2!s}={3!s},{4!s}={5!s}".format(
                       PUSH_ACTION.FIREBASE_CONFIG, self.firebase_config_name,
                       PUSH_ACTION.REGISTRATION_URL, REGISTRATION_URL,
                       PUSH_ACTION.TTL, TTL))

        # 1st step
        with self.app.test_request_context('/token/init',
                                           method='POST',
                                           data={"type": "push",
                                                 "genkey": 1},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            detail = res.json.get("detail")
            serial = detail.get("serial")
            self.assertEqual(detail.get("rollout_state"), "clientwait")
            self.assertTrue("pushurl" in detail)
            # check that the new URL contains the serial number
            self.assertTrue("&serial=PIPU" in detail.get("pushurl").get("value"))
            self.assertFalse("otpkey" in detail)
            enrollment_credential = detail.get("enrollment_credential")

        # 2nd step. Failing with wrong serial number
        with self.app.test_request_context('/ttype/push',
                                           method='POST',
                                           data={"serial": "wrongserial",
                                                 "pubkey": self.smartphone_public_key_pem_urlsafe,
                                                 "fbtoken": "firebaseT"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 404, res)
            status = res.json.get("result").get("status")
            self.assertFalse(status)
            error = res.json.get("result").get("error")
            self.assertEqual(error.get("message"),
                             "No token with this serial number in the rollout state 'clientwait'.")

        # 2nd step. Fails with missing enrollment credential
        with self.app.test_request_context('/ttype/push',
                                           method='POST',
                                           data={"serial": serial,
                                                 "pubkey": self.smartphone_public_key_pem_urlsafe,
                                                 "fbtoken": "firebaseT",
                                                 "enrollment_credential": "WRonG"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 400, res)
            status = res.json.get("result").get("status")
            self.assertFalse(status)
            error = res.json.get("result").get("error")
            self.assertEqual(error.get("message"),
                             "ERR905: Invalid enrollment credential. You are not authorized to finalize this token.")

        # 2nd step: as performed by the smartphone
        with self.app.test_request_context('/ttype/push',
                                           method='POST',
                                           data={"enrollment_credential": enrollment_credential,
                                                 "serial": serial,
                                                 "pubkey": self.smartphone_public_key_pem_urlsafe,
                                                 "fbtoken": "firebaseT"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            detail = res.json.get("detail")
            # still the same serial number
            self.assertEqual(serial, detail.get("serial"))
            self.assertEqual(detail.get("rollout_state"), "enrolled")
            # Now the smartphone gets a public key from the server
            augmented_pubkey = "-----BEGIN RSA PUBLIC KEY-----\n{}\n-----END RSA PUBLIC KEY-----\n".format(
                detail.get("public_key"))
            parsed_server_pubkey = serialization.load_pem_public_key(
                to_bytes(augmented_pubkey),
                default_backend())
            self.assertIsInstance(parsed_server_pubkey, rsa.RSAPublicKey)
            pubkey = detail.get("public_key")

            # Now check, what is in the token in the database
            toks = get_tokens(serial=serial)
            self.assertEqual(len(toks), 1)
            token_obj = toks[0]
            self.assertEqual(token_obj.token.rollout_state, "enrolled")
            self.assertTrue(token_obj.token.active)
            tokeninfo = token_obj.get_tokeninfo()
            self.assertEqual(tokeninfo.get("public_key_smartphone"), self.smartphone_public_key_pem_urlsafe)
            self.assertEqual(tokeninfo.get("firebase_token"), "firebaseT")
            self.assertEqual(tokeninfo.get("public_key_server").strip().strip("-BEGIN END RSA PUBLIC KEY-").strip(),
                             pubkey)
            # The token should also contain the firebase config
            self.assertEqual(tokeninfo.get(PUSH_ACTION.FIREBASE_CONFIG), self.firebase_config_name)
            # remove the token
            remove_token(serial)

    def test_02_api_push_poll(self):
        r = set_smsgateway(self.firebase_config_name,
                           'privacyidea.lib.smsprovider.FirebaseProvider.FirebaseProvider',
                           "myFB", FB_CONFIG_VALS)
        self.assertGreater(r, 0)

        # create a new push token
        tokenobj = self._create_push_token()
        serial = tokenobj.get_serial()

        # set PIN
        tokenobj.set_pin("pushpin")
        tokenobj.add_user(User("cornelius", self.realm1))

        # We mock the ServiceAccountCredentials, since we can not directly contact the Google API
        with mock.patch('privacyidea.lib.smsprovider.FirebaseProvider.service_account.Credentials'
                        '.from_service_account_file') as mySA:
            # alternative: side_effect instead of return_value
            mySA.return_value = _create_credential_mock()

            # add responses, to simulate the communication to firebase
            responses.add(responses.POST, 'https://fcm.googleapis.com/v1/projects/test-123456/messages:send',
                          body="""{}""",
                          content_type="application/json")

            # Send the first authentication request to trigger the challenge.
            # No push notification is submitted to firebase, but a challenge is created anyway
            with mock.patch("logging.Logger.warning") as mock_log:
                with self.app.test_request_context('/validate/check',
                                                   method='POST',
                                                   data={"user": "cornelius",
                                                         "realm": self.realm1,
                                                         "pass": "pushpin"}):
                    res = self.app.full_dispatch_request()
                    self.assertTrue(res.status_code == 200, res)
                    result = res.json.get("result")
                    self.assertTrue(result.get("status"))
                    self.assertFalse(result.get("value"))
                    self.assertEqual("CHALLENGE", result.get("authentication"))
                    # Check that the warning was written to the log file.
                    mock_log.assert_called_with("Failed to submit message to Firebase service for token {0!s}."
                                                .format(serial))

        # first create a signature
        ts = datetime.utcnow().isoformat()
        sign_string = "{serial}|{timestamp}".format(serial=serial, timestamp=ts)
        sig = self.smartphone_private_key.sign(sign_string.encode('utf8'),
                                               padding.PKCS1v15(),
                                               hashes.SHA256())
        # now check that we receive the challenge when polling
        with self.app.test_request_context('/ttype/push',
                                           method='GET',
                                           query_string={"serial": serial,
                                                         "timestamp": ts,
                                                         "signature": b32encode(sig)}):
            res = self.app.full_dispatch_request()

            # check that the serial was set in flask g (via before_request in ttype.py)
            self.assertTrue(self.app_context.g.serial, serial)
            self.assertTrue(res.status_code == 200, res)
            self.assertTrue(res.json['result']['status'])
            chall = res.json['result']['value'][0]
            self.assertTrue(chall)

            challenge = chall["nonce"]
            # This is what the smartphone answers.
            # create the signature:
            sign_data = "{0!s}|{1!s}".format(challenge, serial)
            signature = b32encode_and_unicode(
                self.smartphone_private_key.sign(sign_data.encode("utf-8"),
                                                 padding.PKCS1v15(),
                                                 hashes.SHA256()))

            # Answer the challenge
            with self.app.test_request_context('/ttype/push',
                                               method='POST',
                                               data={"serial": serial,
                                                     "nonce": challenge,
                                                     "signature": signature}):
                res = self.app.full_dispatch_request()
                self.assertTrue(res.status_code == 200, res)
                self.assertTrue(res.json['result']['status'])
                self.assertTrue(res.json['result']['value'])

    def test_03_api_enroll_push_poll_only(self):
        """Enroll a poll-only push token"""
        self.authenticate()
        # Set policy for poll only
        set_policy("push1", scope=SCOPE.ENROLL,
                   action="{0!s}={1!s},{2!s}={3!s},{4!s}={5!s}".format(
                       PUSH_ACTION.FIREBASE_CONFIG, POLL_ONLY,
                       PUSH_ACTION.REGISTRATION_URL, REGISTRATION_URL,
                       PUSH_ACTION.TTL, TTL))

        # 1st step
        with self.app.test_request_context('/token/init',
                                           method='POST',
                                           data={"type": "push",
                                                 "genkey": 1},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            detail = res.json.get("detail")
            serial = detail.get("serial")
            self.assertEqual(detail.get("rollout_state"), "clientwait")
            self.assertIn("pushurl", detail)
            # check that the new URL contains the serial number
            self.assertIn("&serial=PIPU", detail.get("pushurl").get("value"))
            # The firebase settings are NOT contained in the QR Code, since we do poll_only
            # poll_only
            self.assertNotIn("appid=", detail.get("pushurl").get("value"))
            self.assertNotIn("appidios=", detail.get("pushurl").get("value"))
            self.assertNotIn("apikeyios=", detail.get("pushurl").get("value"))
            self.assertNotIn("otpkey", detail)
            enrollment_credential = detail.get("enrollment_credential")

        # 2nd step: as performed by the smartphone. Also in POLL_ONLY the smartphone needs to send
        #           an empty "fbtoken"
        with self.app.test_request_context('/ttype/push',
                                           method='POST',
                                           data={"enrollment_credential": enrollment_credential,
                                                 "serial": serial,
                                                 "pubkey": self.smartphone_public_key_pem_urlsafe,
                                                 "fbtoken": ""}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            detail = res.json.get("detail")
            # still the same serial number
            self.assertEqual(serial, detail.get("serial"))
            self.assertEqual(detail.get("rollout_state"), "enrolled")
            # Now the smartphone gets a public key from the server
            augmented_pubkey = "-----BEGIN RSA PUBLIC KEY-----\n{}\n-----END RSA PUBLIC KEY-----\n".format(
                detail.get("public_key"))
            parsed_server_pubkey = serialization.load_pem_public_key(
                to_bytes(augmented_pubkey),
                default_backend())
            self.assertIsInstance(parsed_server_pubkey, rsa.RSAPublicKey)
            pubkey = detail.get("public_key")

            # Now check, what is in the token in the database
            toks = get_tokens(serial=serial)
            self.assertEqual(len(toks), 1)
            token_obj = toks[0]
            self.assertEqual(token_obj.token.rollout_state, "enrolled")
            self.assertTrue(token_obj.token.active)
            tokeninfo = token_obj.get_tokeninfo()
            self.assertEqual(tokeninfo.get("public_key_smartphone"), self.smartphone_public_key_pem_urlsafe)
            self.assertEqual(tokeninfo.get("firebase_token"), "")
            self.assertEqual(tokeninfo.get("public_key_server").strip().strip("-BEGIN END RSA PUBLIC KEY-").strip(),
                             pubkey)
            # The token should also contain the firebase config
            self.assertEqual(tokeninfo.get(PUSH_ACTION.FIREBASE_CONFIG), POLL_ONLY)

        # remove the token
        remove_token(serial)
        # remove the policy
        delete_policy("push1")

    def test_04_api_poll_declined_chal(self):
        self.setUp_user_realms()
        # create FireBase Service and policies
        set_smsgateway(self.firebase_config_name,
                       'privacyidea.lib.smsprovider.FirebaseProvider.FirebaseProvider',
                       "myFB", FB_CONFIG_VALS)
        set_policy("push_config", scope=SCOPE.ENROLL,
                   action=f"{PUSH_ACTION.FIREBASE_CONFIG}={self.firebase_config_name}")
        set_policy("PUSH_ACTION.REGISTRATION_URL", scope=SCOPE.ENROLL,
                   action=f"{PUSH_ACTION.REGISTRATION_URL}={REGISTRATION_URL}")
        # create push token
        tokenobj = self._create_push_token()
        serial = tokenobj.get_serial()

        # set PIN
        tokenobj.set_pin("pushpin")
        tokenobj.add_user(User("cornelius", self.realm1))

        # We mock the ServiceAccountCredentials, since we can not directly contact the Google API
        with mock.patch('privacyidea.lib.smsprovider.FirebaseProvider.service_account.Credentials'
                        '.from_service_account_file') as mySA:
            # alternative: side_effect instead of return_value
            mySA.from_json_keyfile_name.return_value = _create_credential_mock()

            # add responses, to simulate the communication to firebase
            responses.add_callback(responses.POST, 'https://fcm.googleapis.com/v1/projects/test-123456/messages:send',
                                   callback=_check_firebase_params,
                                   content_type="application/json")

            # Send the first authentication request to trigger the challenge
            with self.app.test_request_context('/validate/check',
                                               method='POST',
                                               data={"user": "cornelius",
                                                     "realm": self.realm1,
                                                     "pass": "pushpin"}):
                res = self.app.full_dispatch_request()
                self.assertTrue(res.status_code == 200, res)
                jsonresp = res.json
                self.assertFalse(jsonresp.get("result").get("value"))
                self.assertTrue(jsonresp.get("result").get("status"))
                self.assertEqual(jsonresp.get("detail").get("serial"), tokenobj.token.serial)
                self.assertTrue("transaction_id" in jsonresp.get("detail"))
                transaction_id = jsonresp.get("detail").get("transaction_id")

            # Our ServiceAccountCredentials mock has not been called because we use a cached token
            self.assertEqual(len(mySA.from_json_keyfile_name.mock_calls), 0)
            self.assertIn(FIREBASE_FILE, get_app_local_store()["firebase_token"])

        # Check if the challenge is sent to the smartphone
        # So when we check later, if the challenge is still sent, we can be sure, that the
        # challenge was not answered.
        timestamp = datetime.utcnow().isoformat()
        sign_string = f"{tokenobj.token.serial}|{timestamp}"
        sig = self.smartphone_private_key.sign(sign_string.encode('utf8'),
                                               padding.PKCS1v15(),
                                               hashes.SHA256())
        with self.app.test_request_context('/ttype/push',
                                           method='GET',
                                           query_string={"serial": tokenobj.token.serial,
                                                         "timestamp": timestamp,
                                                         "signature": b32encode(sig)}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            value = result.get("value")
            self.assertEqual(len(value), 1, value)

        # Decline the challenge
        challengeobject_list = get_challenges(serial=tokenobj.token.serial,
                                              transaction_id=transaction_id)
        challenge = challengeobject_list[0]
        challenge.set_session(CHALLENGE_SESSION.DECLINED)
        challenge.save()

        # If the challenge is successfully declined, the /ttype/push endpoint
        # will not send the challenge again.
        timestamp = datetime.utcnow().isoformat()
        sign_string = f"{tokenobj.token.serial}|{timestamp}"
        sig = self.smartphone_private_key.sign(sign_string.encode('utf8'),
                                               padding.PKCS1v15(),
                                               hashes.SHA256())
        with self.app.test_request_context('/ttype/push',
                                           method='GET',
                                           query_string={"serial": tokenobj.token.serial,
                                                         "timestamp": timestamp,
                                                         "signature": b32encode(sig)}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            value = result.get("value")
            self.assertEqual(len(value), 0, value)
        remove_token(serial=serial)



# 2020-02-15 Jean-Pierre Höhmann <jean-pierre.hoehmann@netknights.it>
#            Add webAuthn token
# 2018-06-15 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#            Add translation for authentication failure - since
#            this is a message that is displayed in the UI.
# 2016-04-08 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#            Avoid "None" as redundant 2nd argument
# 2015-11-04 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#            Add REMOTE_USER check
# 2015-04-03 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#            Add logout time to response
# 2014-12-15 Cornelius Kölbel, info@privacyidea.org
#            Initial creation
#
# (c) Cornelius Kölbel
# Info: http://www.privacyidea.org
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""This REST API is used to authenticate the users. A user needs to
authenticate when he wants to use the API for administrative tasks like
enrolling a token.

This API must not be confused with the validate API, which is used to check,
if a OTP value is valid. See :ref:`rest_validate`.

Authentication of users and admins is tested in tests/test_api_roles.py

You need to authenticate for all administrative tasks. If you are not
authenticated, the API returns a 401 response.

To authenticate you need to send a POST request to /auth containing username
and password.
"""
from flask import (Blueprint,
                   request,
                   current_app,
                   g)
import jwt
from functools import wraps
from datetime import (datetime,
                      timedelta, timezone)
from privacyidea.lib.error import AuthError, ERROR
from privacyidea.lib.crypto import geturandom, init_hsm
from privacyidea.lib.audit import getAudit
from privacyidea.lib.auth import (check_webui_user, ROLE, verify_db_admin,
                                  db_admin_exist)
from privacyidea.lib.fido2.policy_action import FIDO2PolicyAction
from privacyidea.lib.framework import get_app_config_value
from privacyidea.lib.fido2.challenge import verify_fido2_challenge
from privacyidea.lib.fido2.util import get_fido2_token_by_credential_id
from privacyidea.lib.user import User, split_user, log_used_user
from privacyidea.lib.policy import PolicyClass, REMOTE_USER
from privacyidea.lib.realm import get_default_realm, realm_is_defined
from privacyidea.api.lib.postpolicy import (postpolicy, add_user_detail_to_response, check_tokentype,
                                            check_tokeninfo, check_serial, no_detail_on_success,
                                            get_webui_settings)
from privacyidea.api.lib.prepolicy import (is_remote_user_allowed, prepolicy,
                                           pushtoken_disable_wait, webauthntoken_authz, webauthntoken_request,
                                           fido2_auth, increase_failcounter_on_challenge,
                                           jwt_validity)
from privacyidea.api.lib.utils import (send_result, get_all_params,
                                       verify_auth_token, getParam, get_optional, get_required)
from privacyidea.lib.utils import get_client_ip, hexlify_and_unicode, to_unicode, get_plugin_info_from_useragent
from privacyidea.lib.config import get_from_config, SYSCONF, ensure_no_config_object, get_privacyidea_node
from privacyidea.lib.event import event, EventConfiguration
from privacyidea.lib import _
import logging
import traceback
import threading

log = logging.getLogger(__name__)

jwtauth = Blueprint('jwtauth', __name__)


@jwtauth.before_request
def before_request():
    """
    This is executed before the request
    """
    ensure_no_config_object()
    request.all_data = get_all_params(request)
    privacyidea_server = get_app_config_value("PI_AUDIT_SERVERNAME", get_privacyidea_node(request.host))
    g.policy_object = PolicyClass()
    g.audit_object = getAudit(current_app.config)
    g.event_config = EventConfiguration()
    # access_route contains the ip addresses of all clients, hops and proxies.
    g.client_ip = get_client_ip(request,
                                get_from_config(SYSCONF.OVERRIDECLIENT))
    # Save the HTTP header in the localproxy object
    g.request_headers = request.headers
    g.serial = getParam(request.all_data, "serial", default=None)
    g.audit_object.log({"success": False,
                        "client": g.client_ip,
                        "user_agent": get_plugin_info_from_useragent(request.user_agent.string)[0],
                        "user_agent_version": get_plugin_info_from_useragent(request.user_agent.string)[1],
                        "privacyidea_server": privacyidea_server,
                        "action": f"{request.method!s} {request.url_rule!s}",
                        "action_detail": "",
                        "thread_id": f"{threading.current_thread().ident!s}",
                        "info": ""})

    username = getParam(request.all_data, "username")
    if username:
        # We only fill request.User, if we really have a username.
        # On endpoints like /auth/rights, this is not available
        login_name, realm = split_user(username)
        # overwrite the split realm if we have a realm parameter. Default back to default_realm
        realm = getParam(request.all_data, "realm") or realm or get_default_realm()
        # Prefill the request.User. This is used by some pre-event handlers
        try:
            request.User = User(login_name, realm)
        except Exception as e:
            request.User = None
            log.warning(f"Problem resolving user {login_name} in realm {realm}: {e!s}.")
            log.debug(f"{traceback.format_exc()!s}")


@jwtauth.route('', methods=['POST'])
@prepolicy(increase_failcounter_on_challenge, request=request)
@prepolicy(pushtoken_disable_wait, request)
@prepolicy(webauthntoken_request, request=request)
@prepolicy(webauthntoken_authz, request=request)
@prepolicy(fido2_auth, request=request)
@prepolicy(jwt_validity, request)
@postpolicy(get_webui_settings, request=request)
@postpolicy(no_detail_on_success, request=request)
@postpolicy(add_user_detail_to_response, request=request)
@postpolicy(check_tokentype, request=request)
@postpolicy(check_tokeninfo, request=request)
@postpolicy(check_serial, request=request)
@event("auth", request, g)
def get_auth_token():
    """
    This call verifies the credentials of the user and issues an
    authentication token, that is used for the later API calls. The
    authentication token has a validity, that is usually 1 hour.

    :jsonparam username: The username of the user who wants to authenticate to
        the API.
    :jsonparam password: The password/credentials of the user who wants to
        authenticate to the API.
    :jsonparam realm: The realm where the user will be searched.

    :return: A json response with an authentication token, that needs to be
        used in any further request.

    :status 200: in case of success
    :status 401: if authentication fails

    **Example Authentication Request**:

    .. sourcecode:: http

       POST /auth HTTP/1.1
       Host: example.com
       Accept: application/json

       username=admin
       password=topsecret

    **Example Authentication Response**:

    .. sourcecode:: http

       HTTP/1.0 200 OK
       Content-Length: 354
       Content-Type: application/json

       {
            "id": 1,
            "jsonrpc": "2.0",
            "result": {
                "status": true,
                "value": {
                    "token": "eyJhbGciOiJIUz....jdpn9kIjuGRnGejmbFbM"
                }
            },
            "version": "privacyIDEA unknown"
       }

    **Response for failed authentication**:

    .. sourcecode:: http

       HTTP/1.1 401 UNAUTHORIZED
       Content-Type: application/json
       Content-Length: 203

       {
          "id": 1,
          "jsonrpc": "2.0",
          "result": {
            "error": {
              "code": -401,
              "message": "missing Authorization header"
            },
            "status": false
          },
          "version": "privacyIDEA unknown",
          "config": {
            "logout_time": 30
          }
       }

    """
    jwt_validity_param = get_optional(request.all_data, "jwt_validity")
    validity = timedelta(seconds=int(jwt_validity_param))
    username = get_optional(request.all_data, "username")
    password = get_optional(request.all_data, "password")
    realm_param = get_optional(request.all_data, "realm")
    details = {}
    # Passkey login
    credential_id = get_optional(request.all_data, "credential_id")
    passkey_login_enabled = get_app_config_value("WEBUI_PASSKEY_LOGIN_ENABLED", True)
    passkey_login_success = False
    if not passkey_login_enabled and credential_id:
        log.debug("WebUI passkey login disabled in pi.cfg!")
        raise AuthError(_("Authentication with passkey disabled."), id=ERROR.AUTHENTICATE_ILLEGAL_METHOD)
    if credential_id and passkey_login_enabled:
        transaction_id: str = get_required(request.all_data, "transaction_id")
        token = get_fido2_token_by_credential_id(credential_id)
        if not token:
            raise AuthError(_("Authentication failure. The passkey is not registered."),
                            id=ERROR.AUTHENTICATE_WRONG_CREDENTIALS)
        if not token.user:
            raise AuthError(_("Authentication failure. Token has no user."),
                            id=ERROR.AUTHENTICATE_MISSING_USERNAME)
        # TODO For the WebUI login, always require user_verification so that it is a 2FA
        request.all_data.update({FIDO2PolicyAction.USER_VERIFICATION_REQUIREMENT: "required"})
        passkey_login_success = verify_fido2_challenge(transaction_id, token, request.all_data) > 0
        if passkey_login_success:
            user = token.user
            login_name = user.login
            realm = user.realm
            username = user.login
        else:
            raise AuthError(_("Authentication failure using passkey."), id=ERROR.AUTHENTICATE_WRONG_CREDENTIALS)
    # End passkey login
    else:
        # The realm parameter has precedence! Check if it exists
        if realm_param and not realm_is_defined(realm_param):
            raise AuthError(_("Authentication failure. Unknown realm:") + f" {realm_param}.",
                            id=ERROR.AUTHENTICATE_WRONG_CREDENTIALS)

        if username is None:
            raise AuthError(_("Authentication failure. Missing Username"), id=ERROR.AUTHENTICATE_MISSING_USERNAME)

        user = request.User
        if not user:
            # The user could not be resolved, but it could still be a local administrator
            login_name, realm = split_user(username)
            realm = (realm_param or realm or get_default_realm()).lower()
            user = User()
        else:
            realm = user.realm
            login_name = user.login

    # Failsafe to have the user attempt in the log, whatever happens
    # This can be overwritten later
    g.audit_object.log({"user": username, "realm": realm})

    secret = current_app.secret_key
    superuser_realms = [x.lower() for x in current_app.config.get("SUPERUSER_REALM", [])]
    # This is the default role for the logged-in user. The role privileges may be risen to "admin"
    role = ROLE.USER
    # The way the user authenticated. This could be:
    # "password" = The admin user DB or the user store
    # "pi" = The admin or the user is authenticated against privacyIDEA
    # "remote_user" = authenticated by webserver
    authtype = "password"
    # Verify the password
    admin_auth = False
    user_auth = False

    if passkey_login_success:
        authtype = "pi"
        # Login is already completed, get the role of the logged-in user
        if db_admin_exist(username):
            role = ROLE.ADMIN
            admin_auth = True
            g.audit_object.log({"success": True, "user": "", "administrator": username, "info": "internal admin"})
            user = User()
        elif user.realm in superuser_realms:
            role = ROLE.ADMIN
            admin_auth = True
            g.audit_object.log({"user": user.login, "realm": user.realm, "info": log_used_user(user)})
        else:
            user_auth = True
            g.audit_object.log({"user": user.login, "realm": user.realm, "info": log_used_user(user)})

    # Check if the remote user is allowed
    elif (request.remote_user == username) and is_remote_user_allowed(request) != REMOTE_USER.DISABLE:
        # Authenticated by the Web Server
        # Check if the username exists
        # 1. in local admins
        # 2. in a realm
        # 2a. is an admin realm
        authtype = "remote_user "
        if db_admin_exist(username):
            role = ROLE.ADMIN
            admin_auth = True
            g.audit_object.log({"success": True, "user": "", "administrator": username, "info": "internal admin"})
            user = User()
        else:
            # Check if the user exists
            g.audit_object.log({"user": user.login, "realm": user.realm, "info": log_used_user(user)})
            if user.exist():
                user_auth = True
                if user.realm in superuser_realms:
                    role = ROLE.ADMIN
                    admin_auth = True

    elif verify_db_admin(username, password):
        role = ROLE.ADMIN
        admin_auth = True
        log.info(f"Local admin '{username}' successfully logged in.")
        # This admin is not in the default realm!
        realm = ""
        user = User()
        g.audit_object.log({"success": True,
                            "user": "",
                            "realm": "",
                            "administrator": username,
                            "info": "internal admin"})

    else:
        # The user could not be identified against the admin database, so we do the rest of the check
        if password is None:
            g.audit_object.add_to_log({"info": 'Missing parameter "password"'}, add_with_comma=True)
        else:
            options = {"g": g, "clientip": g.client_ip}
            for key, value in request.all_data.items():
                if value and key not in ["g", "clientip"]:
                    options[key] = value
            user_auth, role, details = check_webui_user(user, password, options=options,
                                                        superuser_realms=superuser_realms)
            details = details or {}
            serials = (",".join([challenge_info["serial"] for challenge_info in details["multi_challenge"]])
                       if 'multi_challenge' in details else details.get('serial'))
            if db_admin_exist(user.login) and user_auth and realm == get_default_realm():
                # If there is a local admin with the same login name as the user
                # in the default realm, we inform about this in the log file.
                # This condition can only be checked if the user was authenticated as it
                # is the only way to verify if such a user exists.
                log.warning(f"A user '{user.login}' exists as local admin and as user in your default realm!")
            if role == ROLE.ADMIN:
                g.audit_object.log({"user": "",
                                    "administrator": user.login,
                                    "realm": user.realm,
                                    "resolver": user.resolver,
                                    "serial": serials,
                                    "info": f"{log_used_user(user)}|loginmode={details.get('loginmode')}"})
            else:
                g.audit_object.log({"user": user.login,
                                    "realm": user.realm,
                                    "resolver": user.resolver,
                                    "serial": serials,
                                    "info": f"{log_used_user(user)}|loginmode={details.get('loginmode')}"})

            if not user_auth and "multi_challenge" in details and len(details["multi_challenge"]) > 0:
                # Do not return user data in case of a challenge request.
                return send_result(False, rid=2, details=details)

    if not admin_auth and not user_auth:
        raise AuthError(_("Authentication failure. Wrong credentials"), id=ERROR.AUTHENTICATE_WRONG_CREDENTIALS,
                        details=details or {})
    else:
        g.audit_object.log({"success": True})
        request.User = user

    # If the HSM is not ready, we need to create the nonce in another way!
    hsm = init_hsm()
    if hsm.is_ready:
        nonce = geturandom(hex=True)
        # Add the role to the JWT, so that we can verify it internally and use the authtype for access definitions
        rights = g.policy_object.ui_get_rights(role, realm, login_name, g.client_ip)
        menus = g.policy_object.ui_get_main_menus({"username": login_name, "role": role, "realm": realm}, g.client_ip)
    else:
        import os
        nonce = hexlify_and_unicode(os.urandom(20))
        rights = []
        menus = []

    # What is the log level?
    log_level = current_app.config.get("PI_LOGLEVEL", 30)

    token = jwt.encode({"username": login_name,
                        "realm": realm,
                        "nonce": nonce,
                        "role": role,
                        "authtype": authtype,
                        "exp": datetime.now(timezone.utc) + validity,
                        "rights": rights},
                       secret, algorithm='HS256')

    # set the logged-in user for post-policies and post-events
    g.logged_in_user = {"username": login_name, "realm": realm, "role": role}

    # Add the role to the response, so that the WebUI can make decisions
    # based on this (only show self-service, not the admin part)
    return send_result({"token": to_unicode(token),
                        "role": role,
                        "username": login_name,
                        "realm": realm,
                        "log_level": log_level,
                        "rights": rights,
                        "menus": menus,
                        "auth": True},
                       rid=2,
                       details=details)


def admin_required(f):
    """
    This is a decorator for routes, that require to be authenticated.
    """

    @wraps(f)
    def decorated_function(*args, **kwargs):
        check_auth_token(required_role=[ROLE.ADMIN])
        return f(*args, **kwargs)

    return decorated_function


def user_required(f):
    """
    This is a decorator for routes, that require to be authenticated.
    """

    @wraps(f)
    def decorated_function(*args, **kwargs):
        check_auth_token(required_role=["user", "admin"])
        return f(*args, **kwargs)

    return decorated_function


def check_auth_token(required_role=None):
    """
    This checks the authentication token

    You need to pass an authentication header:

        PI-Authorization: <token>

    You can do this using httpie like this:

        http -j POST http://localhost:5000/system/getConfig Authorization:ewrt
    """
    auth_token = request.headers.get('PI-Authorization')
    if not auth_token:
        auth_token = request.headers.get('Authorization')
    r = verify_auth_token(auth_token, required_role)
    g.logged_in_user = {"username": r.get("username"), "realm": r.get("realm"), "role": r.get("role")}


@jwtauth.route('/rights', methods=['GET'])
@user_required
def get_rights():
    """
    This returns the rights of the logged-in user.

    :reqheader Authorization: The authorization token acquired by /auth request
    """
    enroll_types = g.policy_object.ui_get_enroll_tokentypes(g.client_ip, g.logged_in_user)
    g.audit_object.log({"success": True})
    return send_result(enroll_types)



from .base import MyApiTestCase
from privacyidea.lib.policy import set_policy, delete_policy, SCOPE, ACTION


class APISmsGatewayTestCase(MyApiTestCase):

    def test_01_crud_smsgateway(self):

        # list empty sms gateway definitions
        with self.app.test_request_context('/smsgateway/',
                                           method='GET',
                                           headers={'Authorization': self.at}):

            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value"), [])

        # check that we have an entry in the audit log
        audit_entry = self.find_most_recent_audit_entry(action="GET /smsgateway*")
        self.assertEqual(audit_entry['success'], 1, audit_entry)

        # create an sms gateway definition
        param = {
            "name": "myGW",
            "module": "privacyidea.lib.smsprovider.SMSProvider.ISMSProvider",
            "description": "myGateway",
            "option.URL": "http://example.com"
        }
        with self.app.test_request_context('/smsgateway',
                                           data=param,
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value"), 1)

        # check the gateway
        with self.app.test_request_context('/smsgateway/',
                                           method='GET',
                                           headers={
                                               'Authorization': self.at}):

            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            sms_gw = result.get("value")[0]
            self.assertEqual(sms_gw.get("description"), "myGateway")
            self.assertEqual(sms_gw.get("name"), "myGW")
            self.assertEqual(sms_gw.get("id"), 1)
            self.assertEqual(sms_gw.get("providermodule"),
                             "privacyidea.lib.smsprovider.SMSProvider.ISMSProvider")
            self.assertEqual(sms_gw.get("options").get("URL"),
                             "http://example.com")

        # update
        param = {
            "name": "myGW",
            "module": "privacyidea.lib.smsprovider.SMSProvider.ISMSProvider",
            "description": "new description",
            "id": 1
        }

        with self.app.test_request_context('/smsgateway',
                                           data=param,
                                           method='POST',
                                           headers={'Authorization': self.at}):

            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value"), 1)

        # check the gateway
        with self.app.test_request_context('/smsgateway/1',
                                           method='GET',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            sms_gw = result.get("value")[0]
            self.assertEqual(sms_gw.get("description"), "new description")

        # delete gateway
        with self.app.test_request_context('/smsgateway/myGW',
                                           method='DELETE',
                                           headers={
                                               'Authorization': self.at}):

            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value"), 1)

        # list empty gateways
        with self.app.test_request_context('/smsgateway/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value"), [])

    def test_02_test_options(self):

        # create an sms gateway configuration
        param = {
            "name": "myGW",
            "module": "privacyidea.lib.smsprovider.SMSProvider.ISMSProvider",
            "description": "myGateway",
            "option.URL": "http://example.com",
            "header.header1": "headervalue1"
        }
        with self.app.test_request_context('/smsgateway',
                                           data=param,
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertGreaterEqual(result.get("value"), 1, result)
            smsgw_id = result.get("value")

        # add option
        param["option.HTTP_METHOD"] = "POST"
        # add header
        param["header.header2"] = "headervalue2"
        param["id"] = smsgw_id
        with self.app.test_request_context('/smsgateway',
                                           method='POST',
                                           data=param,
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)

        # check options and headers
        with self.app.test_request_context('/smsgateway/',
                                           method='GET',
                                           headers={
                                                   'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            sms_gw = result.get("value")[0]
            self.assertEqual(sms_gw.get("options").get("URL"),
                             "http://example.com")
            self.assertEqual(sms_gw.get("options").get("HTTP_METHOD"), "POST")
            self.assertEqual(sms_gw.get("headers").get("header1"), "headervalue1")
            self.assertEqual(sms_gw.get("headers").get("header2"), "headervalue2")

        # delete option "URL"
        with self.app.test_request_context('/smsgateway/option/{0!s}/option.URL'.format(smsgw_id),
                                           method='DELETE',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)

        # try to delete header "header1" at the wrong endpoint
        with self.app.test_request_context('/smsgateway/option/{0!s}/option.header1'.format(smsgw_id),
                                           method='DELETE',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 404, res)

        # delete header "header1"
        with self.app.test_request_context('/smsgateway/option/{0!s}/header.header1'.format(smsgw_id),
                                           method='DELETE',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)

        # check options
        with self.app.test_request_context('/smsgateway/',
                                           method='GET',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            sms_gw = result.get("value")[0]
            self.assertEqual(sms_gw.get("options").get("URL"), None)
            self.assertEqual(sms_gw.get("options").get("HTTP_METHOD"),
                             "POST")
            self.assertEqual(sms_gw.get("headers").get("header1"), None)
            self.assertEqual(sms_gw.get("headers").get("header2"), "headervalue2")

    def test_04_sms_provider_modules(self):
        with self.app.test_request_context('/smsgateway/providers',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            value = res.json.get("result").get("value")
            self.assertEqual(len(value), 6)
            self.assertTrue('privacyidea.lib.smsprovider.HttpSMSProvider'
                            '.HttpSMSProvider' in value)
            self.assertTrue('privacyidea.lib.smsprovider.SmtpSMSProvider'
                            '.SmtpSMSProvider' in value)
            self.assertTrue('privacyidea.lib.smsprovider.SipgateSMSProvider'
                            '.SipgateSMSProvider' in value)
            self.assertTrue('privacyidea.lib.smsprovider.SmppSMSProvider'
                            '.SmppSMSProvider' in value)
            self.assertIn('privacyidea.lib.smsprovider.ScriptSMSProvider'
                          '.ScriptSMSProvider', value)
            http_parameters = value.get('privacyidea.lib.smsprovider.'
                                        'HttpSMSProvider.HttpSMSProvider')
            smtp_parameters = value.get('privacyidea.lib.smsprovider.'
                                        'SmtpSMSProvider.SmtpSMSProvider')
            sipgate_parameters = value.get('privacyidea.lib.smsprovider.'
                                        'SipgateSMSProvider.SipgateSMSProvider')
            smpp_parameters = value.get('privacyidea.lib.smsprovider.'
                                        'SmppSMSProvider.SmppSMSProvider')
            self.assertEqual(http_parameters.get("options_allowed"), True)
            self.assertEqual(smtp_parameters.get("options_allowed"), False)
            self.assertEqual(sipgate_parameters.get("options_allowed"), False)
            self.assertEqual(smpp_parameters.get("options_allowed"), False)
            self.assertTrue("URL" in http_parameters.get("parameters"))
            self.assertTrue("PROXY" in http_parameters.get("parameters"))
            self.assertTrue("HTTP_METHOD" in http_parameters.get("parameters"))

    def test_05_read_write_policies(self):
        set_policy(name="pol_read", scope=SCOPE.ADMIN,
                   action=ACTION.SMSGATEWAYREAD)
        # create an sms gateway configuration
        param = {
            "name": "myGW",
            "module": "privacyidea.lib.smsprovider.SMSProvider.ISMSProvider",
            "description": "myGateway",
            "option.URL": "http://example.com"
        }
        with self.app.test_request_context('/smsgateway',
                                           data=param,
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 403, res)

        # Now we create a write policy, and we are allowed to write
        set_policy(name="pol_write", scope=SCOPE.ADMIN,
                   action=ACTION.SMSGATEWAYWRITE)
        with self.app.test_request_context('/smsgateway',
                                           data=param,
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)

        # delete the read policy
        delete_policy("pol_read")

        with self.app.test_request_context('/smsgateway/',
                                           method='GET',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 403, res)

        # delete the write policy
        delete_policy("pol_write")

        # and delete sms gateway
        with self.app.test_request_context('/smsgateway/myGW',
                                           method='DELETE',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)



"""
This tests the files
  lib/counter.py
"""
import mock
from contextlib import contextmanager

from .base import MyTestCase
from privacyidea.lib.counter import increase, decrease, reset, read
from privacyidea.models import EventCounter


def increase_and_read(name):
    """ helper function that increases the event counter and returns the new value """
    increase(name)
    return read(name)


def decrease_and_read(name, allow_negative=False):
    """ helper function that decreases the event counter and returns the new value """
    decrease(name, allow_negative=allow_negative)
    return read(name)

class CounterTestCase(MyTestCase):
    """
    Test the counter module
    """

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_00_create_and_increase(self):
        r = increase_and_read("hallo_counter")
        self.assertEqual(r, 1)

        r = increase_and_read("counter2")
        self.assertEqual(r, 1)

        r = increase_and_read("hallo_counter")
        self.assertEqual(r, 2)

        counter = EventCounter.query.filter_by(counter_name="hallo_counter").first()
        self.assertEqual(counter.counter_value, 2)

    def test_01_increase_and_decrease(self):
        r = increase_and_read("hallo_counter1")
        self.assertEqual(r, 1)

        for x in range(1, 5):
            increase("hallo_counter1")

        counter = EventCounter.query.filter_by(counter_name="hallo_counter1").first()
        self.assertEqual(counter.counter_value, 5)

        r = decrease_and_read("hallo_counter1")
        self.assertEqual(r, 4)

        for x in range(1, 8):
            decrease("hallo_counter1")

        counter = EventCounter.query.filter_by(counter_name="hallo_counter1").first()
        self.assertEqual(counter.counter_value, 0)

        # Test reading counter
        r = read("hallo_counter1")
        self.assertEqual(r, 0)
        r = read("unknown counter")
        self.assertEqual(r, None)

    def test_02_decrease_beyond_zero(self):
        r = increase_and_read("hallo_counter2")
        self.assertEqual(r, 1)

        for x in range(1, 8):
            decrease("hallo_counter2", allow_negative=True)

        counter = EventCounter.query.filter_by(counter_name="hallo_counter2").first()
        self.assertEqual(counter.counter_value, -6)

    def test_03_decrease_and_reset(self):
        r = decrease_and_read("hallo_counter3", allow_negative=True)
        self.assertEqual(r, -1)

        reset("hallo_counter3")

        counter = EventCounter.query.filter_by(counter_name="hallo_counter3").first()
        self.assertEqual(counter.counter_value, 0)

    def test_04_reset_non_existing_counter(self):
        reset("hallo_counter4")

        counter = EventCounter.query.filter_by(counter_name="hallo_counter4").first()
        self.assertEqual(counter.counter_value, 0)

    def test_05_multiple_nodes(self):
        @contextmanager
        def _set_node(node):
            """ context manager that sets the current node name """
            with mock.patch("privacyidea.lib.counter.get_privacyidea_node") as mock_node:
                mock_node.return_value = node
                yield

        # two nodes node1 and node2, two counters ctrA and ctrB
        with _set_node("node1"):
            for _ in range(3):
                increase("ctrA")
            increase("ctrB")
        with _set_node("node2"):
            r = increase_and_read("ctrB")
            self.assertEqual(r, 2)

        # sums are correct ...
        self.assertEqual(read("ctrA"), 3)
        self.assertEqual(read("ctrB"), 2)
        # ... and each node has written to its own row
        self.assertEqual(EventCounter.query.filter_by(counter_name="ctrA", node="node1").one().counter_value, 3)
        self.assertEqual(EventCounter.query.filter_by(counter_name="ctrA", node="node2").all(), [])
        self.assertEqual(EventCounter.query.filter_by(counter_name="ctrB", node="node1").one().counter_value, 1)
        self.assertEqual(EventCounter.query.filter_by(counter_name="ctrB", node="node2").one().counter_value, 1)

        # decreasing ctrB on node2 by 2 creates a row with negative value, even if allow_negative=False
        with _set_node("node2"):
            for _ in range(2):
                decrease("ctrB", allow_negative=False)

        self.assertEqual(read("ctrB"), 0)
        self.assertEqual(EventCounter.query.filter_by(counter_name="ctrB", node="node1").one().counter_value, 1)
        self.assertEqual(EventCounter.query.filter_by(counter_name="ctrB", node="node2").one().counter_value, -1)

        # decreasing below the sum of 0 causes all values to be reset
        with _set_node("node2"):
            decrease("ctrB", allow_negative=False)

        self.assertEqual(read("ctrB"), 0)
        self.assertEqual(EventCounter.query.filter_by(counter_name="ctrB", node="node1").one().counter_value, 0)
        self.assertEqual(EventCounter.query.filter_by(counter_name="ctrB", node="node2").one().counter_value, 0)

        # decreasing with allow_negative=True works
        with _set_node("node1"):
            decrease("ctrB", allow_negative=True)
        with _set_node("node2"):
            decrease("ctrB", allow_negative=True)

        self.assertEqual(read("ctrB"), -2)
        self.assertEqual(EventCounter.query.filter_by(counter_name="ctrB", node="node1").one().counter_value, -1)
        self.assertEqual(EventCounter.query.filter_by(counter_name="ctrB", node="node2").one().counter_value, -1)

        # resetting resets all rows
        reset("ctrB")
        self.assertEqual(read("ctrB"), 0)
        self.assertEqual(EventCounter.query.filter_by(counter_name="ctrB", node="node1").one().counter_value, 0)
        self.assertEqual(EventCounter.query.filter_by(counter_name="ctrB", node="node2").one().counter_value, 0)


# SPDX-FileCopyrightText: (C) 2023 Paul Lettich <paul.lettich@netknights.it>
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Info: https://privacyidea.org
#
# This code is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation, either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program. If not, see <http://www.gnu.org/licenses/>.
"""CLI commands for managing tokens"""
import click
from flask.cli import AppGroup
from privacyidea.lib.token import import_token
from privacyidea.lib.importotp import parseOATHcsv


token_cli = AppGroup("token", short_help="Manage tokens in privacyIDEA",
                     help="Commands to manage token in privacyIDEA")


@token_cli.command("import", short_help="Import tokens from a file")
@click.argument("file", type=click.File())
@click.option("-t", "--tokenrealm", multiple=True, default=[],
              help="The realms in which the tokens should be imported (can be "
                   "used multiple times)")
def import_tokens(file, tokenrealm):
    """
    Import Tokens from CSV data in FILE
    """
    contents = file.read()
    tokens = parseOATHcsv(contents)
    i = 0
    for serial in tokens:
        i += 1
        print(u"{0!s}/{1!s} Importing token {2!s}".format(i, len(tokens), serial))

        import_token(serial, tokens[serial], tokenrealms=tokenrealm)



#  2018-06-25 Friedrich Weber <friedrich.weber@netknights.it>
#             Initial implementation of periodic tasks
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

__doc__ = """This module provides functions to manage periodic tasks in the database,
to determine their next scheduled running time and to run them."""

import logging
from datetime import datetime
from privacyidea.lib.tokenclass import DATE_FORMAT

from croniter import croniter
from dateutil.tz import tzutc, tzlocal

from privacyidea.lib.error import ParameterError, ResourceNotFoundError
from privacyidea.lib.utils import fetch_one_resource, parse_date
from privacyidea.lib.task.eventcounter import EventCounterTask
from privacyidea.lib.task.simplestats import SimpleStatsTask
from privacyidea.models import PeriodicTask
from privacyidea.lib.framework import get_app_config
from privacyidea.lib.utils.export import (register_import, register_export)

log = logging.getLogger(__name__)

TASK_CLASSES = [EventCounterTask, SimpleStatsTask]
#: TASK_MODULES maps task module identifiers to subclasses of BaseTask
TASK_MODULES = dict((cls.identifier, cls) for cls in TASK_CLASSES)


def get_available_taskmodules():
    """
    Return a list of all available task module identifiers.
    :return: a list of strings
    """
    return list(TASK_MODULES.keys())


def get_taskmodule(identifier, config=None):
    """
    Return an instance of the given task module. Raise ParameterError if it does not exist.
    :param identifier: identifier of the task module
    :return: instance of a BaseTask subclass
    """
    if identifier not in TASK_MODULES:
        raise ParameterError("Unknown task module: {!r}".format(identifier))
    else:
        r = TASK_MODULES[identifier](config=get_app_config())
        return r


def calculate_next_timestamp(ptask, node, interval_tzinfo=None):
    """
    Calculate the timestamp of the next scheduled run of task ``ptask`` on node ``node``.
    We do not check if the task is even scheduled to run on the specified node.
    Malformed cron expressions may throw a ``ValueError``.

    The next timestamp is calculated based on the last time the task was run on the given node.
    If the task has never run on the node, the last update timestamp of the periodic tasks
    is used as a reference timestamp.

    :param ptask: Dictionary describing the periodic task, as from ``PeriodicTask.get()``
    :param node: Node on which the periodic task is scheduled
    :type node: unicode
    :param interval_tzinfo: Timezone in which the cron expression should be interpreted. Defaults to local time.
    :type interval_tzinfo: tzinfo
    :return: a timezone-aware (UTC) datetime object
    """
    if interval_tzinfo is None:
        interval_tzinfo = tzlocal()
    timestamp = ptask["last_runs"].get(node, ptask["last_update"])
    local_timestamp = timestamp.astimezone(interval_tzinfo)
    iterator = croniter(ptask["interval"], local_timestamp)
    next_timestamp = iterator.get_next(datetime)
    # This will again be a timezone-aware datetime, but we return a timezone-aware UTC timestamp
    return next_timestamp.astimezone(tzutc())


def set_periodic_task(name=None, interval=None, nodes=None, taskmodule=None,
                      ordering=0, options=None, active=True, id=None,
                      retry_if_failed=True):
    """
    Set a periodic task configuration. If ``id`` is None, this creates a new database entry.
    Otherwise, an existing entry is overwritten. We actually ensure that such
    an entry exists and throw a ``ParameterError`` otherwise.

    This also checks if ``interval`` is a valid cron expression, and throws
    a ``ParameterError`` if it is not.

    :param name: Unique name of the periodic task
    :type name: unicode
    :param interval: Periodicity as a string in crontab format
    :type interval: unicode
    :param nodes: List of nodes on which this task should be run
    :type nodes: list of unicode
    :param taskmodule: Name of the task module
    :type taskmodule: unicode
    :param ordering: Ordering of the periodic task (>= 0). Lower numbers are executed first.
    :type ordering: int
    :param options: Additional options for the task module
    :type options: Dictionary mapping unicodes to values that can be converted to unicode or None
    :param active: Flag determining whether the periodic task is active
    :type active: bool
    :param retry_if_failed: true if privacyidea should retry to execute this periodic task if it fails
                            false if privacyidea should just try onetime regardless the failing of the task
    :type retry_if_failed: bool
    :param id: ID of the existing entry, or None
    :type id: int or None
    :return: ID of the entry
    """
    try:
        croniter(interval)
    except ValueError as e:
        raise ParameterError("Invalid interval: {!s}".format(e))
    if ordering < 0:
        raise ParameterError("Invalid ordering: {!s}".format(ordering))
    if id is not None:
        # This will throw a ParameterError if there is no such entry
        get_periodic_task_by_id(id)
    periodic_task = PeriodicTask(name, active, interval, nodes, taskmodule,
                                 ordering, options, id, retry_if_failed)
    return periodic_task.id


def delete_periodic_task(ptask_id):
    """
    Delete an existing periodic task. If ``ptask_id`` refers to an unknown entry, a ParameterError is raised.
    :param ptask_id: ID of the database entry
    :return: ID of the deleted entry
    """
    periodic_task = _get_periodic_task_entry(ptask_id)
    return periodic_task.delete()


def enable_periodic_task(ptask_id, enable=True):
    """
    Set the ``active`` flag of an existing periodic task to ``enable``.
    If ``ptask_id`` refers to an unknown entry, a ParameterError is raised.
    :param ptask_id: ID of the database entry
    :param enable: New value of the ``active`` flag
    :return: ID of the database entry
    """
    periodic_task = _get_periodic_task_entry(ptask_id)
    periodic_task.active = enable
    return periodic_task.save()


def get_periodic_tasks(name=None, node=None, active=None):
    """
    Get a list of all periodic tasks, or of all tasks satisfying a filter criterion,
    ordered by their ordering value (ascending).

    :param name: Name of the periodic task
    :type name: unicode
    :param node: Node for which periodic tasks should be collected. This only includes
                 periodic tasks which are scheduled to run on ``node``.
    :type node: unicode
    :param active: This can be used to filter for active or inactive tasks only
    :return: A (possibly empty) list of periodic task dictionaries
    """
    query = PeriodicTask.query
    if name is not None:
        query = query.filter_by(name=name)
    if active is not None:
        query = query.filter_by(active=active)
    entries = query.order_by(PeriodicTask.ordering).all()
    result = []
    for entry in entries:
        ptask = entry.get()
        if node is None or node in ptask["nodes"]:
            result.append(ptask)
    return result


def get_periodic_task_by_name(name):
    """
    Get a periodic task by name. Raise ParameterError if the task could not be found.
    :param name: task name, unicode
    :return: dictionary
    """
    periodic_tasks = get_periodic_tasks(name)
    if len(periodic_tasks) != 1:
        raise ResourceNotFoundError("The periodic task with unique name {!r} does not exist".format(name))
    return periodic_tasks[0]


def get_periodic_task_by_id(ptask_id):
    """
    Get a periodic task entry by ID and return it as a dictionary.
    Raise ParameterError if the task could not be found.
    :param ptask_id: task ID as integer
    :return: dictionary
    """
    return _get_periodic_task_entry(ptask_id).get()


def _get_periodic_task_entry(ptask_id):
    """
    Get a periodic task entry by ID. Raise ResourceNotFoundError if the task could not be found.
    This is only for internal use.
    :param id: task ID as integer
    :return: PeriodicTask object
    """
    return fetch_one_resource(PeriodicTask, id=ptask_id)


def set_periodic_task_last_run(ptask_id, node, last_run_timestamp):
    """
    Write to the database the information that the specified
    periodic task has been run on a node at a given time.
    :param ptask_id: ID of the periodic task. Raises ParameterError if unknown.
    :type ptask_id: int
    :param node: Node name. It is not checked whether the task is scheduled to run on that node!
    :type node: unioode
    :param last_run_timestamp: Timestamp of the last run
    :type last_run_timestamp: timezone-aware datetime object
    """
    periodic_task = _get_periodic_task_entry(ptask_id)
    utc_last_run = last_run_timestamp.astimezone(tzutc()).replace(tzinfo=None)
    periodic_task.set_last_run(node, utc_last_run)


def get_scheduled_periodic_tasks(node, current_timestamp=None, interval_tzinfo=None):
    """
    Collect all periodic tasks that should be run on a specific node, ordered by
    their ordering.

    This function is usually called by the local cron runner which is aware of the
    current local node name.

    :param node: Node name
    :type node: unicode
    :param current_timestamp: The current timestamp, defaults to the current time
    :type current_timestamp: timezone-aware datetime
    :param interval_tzinfo: timezone in which the crontab expression should be interpreted
    :type interval_tzinfo: tzinfo, defaults to local time
    :return: List of periodic task dictionaries
    """
    active_ptasks = get_periodic_tasks(node=node, active=True)
    if current_timestamp is None:
        current_timestamp = datetime.now(tzutc())
    if current_timestamp.tzinfo is None:
        raise ParameterError("expected timezone-aware datetime, got {!r}".format(current_timestamp))
    scheduled_ptasks = []
    log.debug("Collecting periodic tasks to run at {!s}".format(current_timestamp.isoformat()))
    for ptask in active_ptasks:
        try:
            next_timestamp = calculate_next_timestamp(ptask, node, interval_tzinfo)
            log.debug("Next scheduled run of {!r}: {!s}".format(ptask["name"], next_timestamp.isoformat()))
            if next_timestamp <= current_timestamp:
                log.debug("Scheduling periodic task {!r}".format(ptask["name"]))
                scheduled_ptasks.append(ptask)
        except Exception as e:
            log.warning("Ignoring periodic task {!r}: {!r}".format(ptask["name"], e))
    return scheduled_ptasks


def execute_task(taskmodule, params):
    """
    Given a task module name, run the task with the given parameters.
    :param config: The app configuration
    :param taskmodule: unicode determining the task module
    :param params: dictionary mapping task option keys (unicodes) to unicodes (or None)
    :return: boolean returned by the task
    """
    module = get_taskmodule(taskmodule)
    log.info("Running taskmodule {!r} with parameters {!r}".format(module, params))
    return module.do(params)


@register_export('periodictask')
def export_periodictask(name=None):
    """ Export given or all machineresolver configuration """
    periodictask_list = get_periodic_tasks(name=name)
    # last_update is not required for import, but used here for completeness
    for task in periodictask_list:
        last_update = task.get('last_update')
        task['last_update'] = last_update.strftime(DATE_FORMAT)

    return periodictask_list


@register_import('periodictask')
def import_periodictask(data, name=None):
    """Import periodictask configuration"""
    log.debug('Import periodictask config: {0!s}'.format(data))
    for res_data in data:
        if name and name != res_data.get('name'):
            continue
        res_data.pop('last_update')
        res_data.pop('last_runs')
        rid = set_periodic_task(**res_data)
        log.info('Import of periodictask "{0!s}" finished,'
                 ' id: {1!s}'.format(res_data['name'], rid))



#!/usr/bin/env python

import base64
import getpass
import hashlib
import os


def get_hash(passwd, salt):
    """
    Calculate a sha1 hash of a password with salt.

    :param passwd: The password
    :type passwd: string
    :param salt: The salt
    :type salt: bytestring
    :return: string

    let's see if doctest works:

    >>> print(get_hash('test', b'test'))
    {SSHA}Uau5Y2B43vv4iNhFenx2+FyPEUx0ZXN0

    """
    h = hashlib.sha1(passwd.encode('utf8') + salt).digest()
    return '{SSHA}' + base64.b64encode(h + salt).decode('utf8')


if __name__ == '__main__':
    salt = os.urandom(8)  # edit the length as you see fit
    print(get_hash(getpass.getpass(), salt))



class FIDO2TokenInfo:
    PUB_KEY = "pubKey"
    ORIGIN = "origin"
    AAGUID = "aaguid"
    ATTESTATION_LEVEL = "attestation_level"
    ATTESTATION_ISSUER = "attestation_issuer"
    ATTESTATION_SERIAL = "attestation_serial"
    ATTESTATION_SUBJECT = "attestation_subject"
    RELYING_PARTY_ID = "relying_party_id"
    RELYING_PARTY_NAME = "relying_party_name"
    CREDENTIAL_ID_HASH = "credential_id_hash"
    SIGN_COUNT = "sign_count"
    DEVICE_TYPE = "device_type"
    BACKED_UP = "backed_up"
    PUBLIC_KEY = "public_key"
    ATTESTATION_CERTIFICATE = "attestation_certificate"
    USER_ID = "fido2_user_id"
    CREDENTIAL_ID_HASH = "credential_id_hash"
    SIGN_COUNT = "sign_count"
    DEVICE_TYPE = "device_type"
    BACKED_UP = "backed_up"
    PUBLIC_KEY = "public_key"
    ATTESTATION_CERTIFICATE = "attestation_certificate"
    USER_ID = "fido2_user_id"


"""
This test file tests the lib.auth_cache.py

The lib.auth_cache.py only depends on the database model.
"""
from .base import MyTestCase

from privacyidea.lib.authcache import (add_to_cache, delete_from_cache,
                                       update_cache, verify_in_cache,
                                       _hash_password,
                                       cleanup)
from passlib.hash import argon2
from privacyidea.models import AuthCache
import datetime


class AuthCacheTestCase(MyTestCase):
    """
    Test the policies on a database level
    """
    password = "secret123456"
    username = "hans"
    realm = "realm"
    resolver = "resolver"

    def test_01_write_update_delete_cache(self):
        teststart = datetime.datetime.utcnow()

        r = add_to_cache(self.username, self.realm, self.resolver, self.password)

        self.assertTrue(r > 0)

        auth = AuthCache.query.filter(AuthCache.id == r).first()
        self.assertEqual(auth.username, self.username)
        self.assertTrue(argon2.verify(self.password, auth.authentication))

        self.assertTrue(auth.first_auth > teststart)
        self.assertEqual(auth.last_auth, auth.first_auth)

        update_cache(r)
        auth = AuthCache.query.filter(AuthCache.id == r).first()
        self.assertTrue(auth.last_auth > teststart)

        r_delete = delete_from_cache(self.username, self.realm, self.resolver,
                                     self.password)
        self.assertEqual(r, r_delete)

        auth = AuthCache.query.filter(AuthCache.username ==
                                      self.username).first()
        self.assertEqual(auth, None)

    def test_02_verify_cache(self):
        # Verify if no entry in cache
        r = verify_in_cache(self.username, self.realm, self.resolver,
                            self.password)
        self.assertFalse(r)

        # Add Entry to cache
        ac_id = add_to_cache(self.username, self.realm, self.resolver, self.password)

        auth = AuthCache.query.filter(AuthCache.id == ac_id).first()
        last_auth1 = auth.last_auth
        auth_count1 = auth.auth_count
        self.assertEqual(0, auth_count1)

        first_auth = datetime.datetime.utcnow() - datetime.timedelta(hours=4)
        last_auth = datetime.datetime.utcnow() - datetime.timedelta(minutes=5)

        r = verify_in_cache(self.username, self.realm, self.resolver,
                            self.password, first_auth=first_auth,
                            last_auth=last_auth, max_auths=1)
        self.assertTrue(r)

        # The last_auth was increased!
        auth = AuthCache.query.filter(AuthCache.id == ac_id).first()
        last_auth2 = auth.last_auth
        auth_count2 = auth.auth_count
        self.assertTrue(last_auth2 > last_auth1)
        self.assertEqual(1, auth_count2)

        r = verify_in_cache(self.username, self.realm, self.resolver,
                            self.password, first_auth=first_auth,
                            last_auth=last_auth, max_auths=1)
        self.assertFalse(r)

    def test_03_delete_old_entries(self):
        # Create a VERY old authcache entry
        AuthCache("grandpa", self.realm, self.resolver, _hash_password(self.password),
                  first_auth=datetime.datetime.utcnow() - datetime.timedelta(
                      days=10),
                  last_auth=datetime.datetime.utcnow() - datetime.timedelta(
                      days=2)).save()

        r = verify_in_cache(
            "grandpa", self.realm, self.resolver,
            self.password,
            first_auth=datetime.datetime.utcnow() - datetime.timedelta(hours=4),
            last_auth=datetime.datetime.utcnow() - datetime.timedelta(
                minutes=5))
        # Verification for grandpa fails.
        self.assertFalse(r)
        # The verify_in_cache should have deleted old grandpa entries:
        r = AuthCache.query.filter(AuthCache.username == "grandpa").first()
        self.assertEqual(r, None)

    def test_04_cleanup_authcache(self):
        # cleanup everything!
        r = cleanup(100000000)
        # Create some entries:
        AuthCache("grandpa", self.realm, self.resolver, _hash_password(self.password),
                  first_auth=datetime.datetime.utcnow() - datetime.timedelta(
                      days=10),
                  last_auth=datetime.datetime.utcnow() - datetime.timedelta(
                      days=2)).save()
        AuthCache("grandpa", self.realm, self.resolver, _hash_password(self.password),
                  first_auth=datetime.datetime.utcnow() - datetime.timedelta(
                      minutes=10),
                  last_auth=datetime.datetime.utcnow() - datetime.timedelta(
                      minutes=2)).save()

        # Now we delete entries, that are older than 20 minutes. Only the 2 days old
        # should be deleted. Not the 2 minutes old.
        r = cleanup(10)
        self.assertEqual(1, r)

    def test_05_old_hashes(self):
        from privacyidea.lib.crypto import hash
        # Test that old hashes do not break the code
        r = cleanup(100000000)
        # Add an entry with an old password hash
        AuthCache("grandpa", self.realm, self.resolver, hash("old password", seed=""),
                  first_auth=datetime.datetime.utcnow() - datetime.timedelta(
                      minutes=10),
                  last_auth=datetime.datetime.utcnow() - datetime.timedelta(
                      minutes=2)).save()

        r = verify_in_cache("grandpa", self.realm, self.resolver, "old password")
        self.assertFalse(r)

    def test_06_delete_other_invalid_entries(self):
        # Test deletion of expired entries
        r1 = add_to_cache(self.username, self.realm, self.resolver, "somethingDifferent")
        r2 = add_to_cache(self.username, self.realm, self.resolver, self.password)

        auth1 = AuthCache.query.filter(AuthCache.id == r1).first()
        auth2 = AuthCache.query.filter(AuthCache.id == r2).first()
        last_valid_cache_time = auth1.first_auth
        new_valid_cache_time = auth2.first_auth

        self.assertFalse(auth1.first_auth == auth2.first_auth)

        # delete entries where password matches or first_auth is older than last_valid_cache_time
        delete_from_cache(self.username, self.realm, self.resolver, self.password,
                          last_valid_cache_time=last_valid_cache_time)

        auth = AuthCache.query.filter(AuthCache.username == self.username).first()
        # r2 should have been deleted since the password matches,
        # r1 is still there since it's first_auth is equal to last_valid_cache_time
        self.assertEqual(auth.id, r1)

        # by setting the last_valid_cache_time to the first_auth of r2, the
        # entry r1 should be deleted as well
        delete_from_cache(self.username, self.realm, self.resolver, 'unknown_pw',
                          last_valid_cache_time=new_valid_cache_time)

        auth = AuthCache.query.filter(AuthCache.username == self.username).first()
        self.assertEqual(None, auth)

        # Test deletion if max_auths is reached
        r1 = add_to_cache(self.username, self.realm, self.resolver, self.password)
        r2 = add_to_cache(self.username, self.realm, self.resolver, "somethingDifferent")

        update_cache(r2)
        update_cache(r2)

        auth2 = AuthCache.query.filter(AuthCache.id == r2).first()

        self.assertEqual(2, auth2.auth_count)
        # this deletes the entries matching the password and max_auth
        delete_from_cache(self.username, self.realm, self.resolver, self.password, max_auths=2)

        auth = AuthCache.query.filter(AuthCache.username == self.username).first()
        self.assertEqual(auth, None)



"""merge two heads of audit log and auth_cache

Revision ID: d5870fd2f2a4
Revises: ('3236a1abf1c6', '8d40dbcfda25')
Create Date: 2020-11-21 14:12:41.097913

"""

# revision identifiers, used by Alembic.
revision = 'd5870fd2f2a4'
down_revision = ('3236a1abf1c6', '8d40dbcfda25')

from alembic import op
import sqlalchemy as sa


def upgrade():
    pass


def downgrade():
    pass



#  2022-03-10 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Init

#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
import os
import time
import sys
import contextlib
import getopt
import logging
from privacyidea.lib.security.default import (DefaultSecurityModule,
                                              int_list_to_bytestring)
from privacyidea.lib.error import HSMException
from getpass import getpass


__doc__ = """
This is a PKCS11 Security module that decrypts a given encrypted key file
in the HSM with an asymmetric private key.

Please see the docstring of the SecurityModule below, on how to configure this
module in pi.cfg.

If you have a plain text 96 bytes key file, you can encrypt this file with a given key id like this::

  python encryptkey.py --module /usr/lib/libykcs11.so --keyid 1 --slotname "Yubico YubiKey" \
                       --infile enckey --outfile enckey.enc

Some HSMs have not working key ids and rather use key labels to identify the keys. Then you can encrypt
an existing key file like this::

  python encryptkey.py --module /usr/lib/libykcs11.so --keylabel "my secret key" --slotname "Yubico YubiKey" \
                       --infile enckey --outfile enckey.enc
"""

log = logging.getLogger(__name__)

try:
    import PyKCS11
    MECHANISM = PyKCS11.CKM_RSA_PKCS
except ImportError:
    log.info("The python module PyKCS11 is not available. "
             "So we can not use the PKCS11 security module.")

# The lock directory is used for locking the different processes during startup
# to avoid a deadlock when accessing the HSM.
# The directory for logging can be configured via PI_HSM_MODULE_LOCK_DIR in pi.cfg
DEFAULT_LOCK_DIR = "/dev/shm/pilock"  # nosec B108 # Used for locking during startup
DEFAULT_TIMEOUT = 15


@contextlib.contextmanager
def hsm_lock(timeout=DEFAULT_TIMEOUT, lock_dir=DEFAULT_LOCK_DIR):
    # Wait, that we are free to go
    i_created_the_lock = False
    while timeout > 0:
        timeout -= 1
        try:
            # Try to get the Lock, since we are acting.
            log.debug("Requesting lock")
            os.mkdir(lock_dir)
            log.debug("Lock acquired")
            i_created_the_lock = True
            yield
            break
        except FileExistsError:
            # Some other process got the lock in the meantime.
            log.info("Can not get the lock on {0!s}. Can not initialize the HSM, yet.".format(lock_dir))
            time.sleep(1)
        finally:
            # Cleanup
            if i_created_the_lock:
                log.debug("Going to release lock")
                os.rmdir(lock_dir)
                log.debug("Lock released")


class EncryptKeyHardwareSecurityModule(DefaultSecurityModule):  # pragma: no cover

    def __init__(self, config=None, logout=True):
        """
        Initialize the PKCS11 Security Module.
        The configuration needs to contain the pkcs11 module and the ID of the key.

        {"module": "/usr/lib/libykcs11.so",
         "slotname": "Yubico YubiKey",
         "keyid": 1,
         "keylabel": "my secret key"
         "encfile": "/etc/privacyidea/enckey.enc",
         "password", "123456"}

        The encfile is the encrypted encryption key.

        The values can be configured in the pi.cfg file like this:

        PI_HSM_MODULE = "privacyidea.lib.security.encryptkey.EncryptKeyHardwareSecurityModule"
        PI_HSM_MODULE_MODULE = "/usr/lib/libykcs11.so"
        PI_HSM_MODULE_SLOTNAME = "Yubico YubiKey"
        # Alternative to slotname:
        # PI_HSM_MODULE_SLOT = 1
        PI_HSM_MODULE_KEYID = 1
        # Alternative to KEYID -- only use one of both!
        PI_HSM_MODULE_KEYLABEL = "my secret key"
        PI_HSM_MODULE_PASSWORD = '123456'
        PI_HSM_MODULE_ENCFILE = "/etc/privacyidea/enckey.enc"

        :param config: contains the HSM configuration
        :type config: dict

        :return: The Security Module object
        """
        self.config = config or {}
        self.secrets = {}
        self.name = "HSM"
        self.slot = self.config.get("slot") or -1
        self.slotname = self.config.get("slotname")
        if "module" not in config:
            log.error("No PKCS11 module defined!")
            raise HSMException("No PKCS11 module defined.")
        else:
            self.module = self.config.get("module")

        if "keyid" not in config and "keylabel" not in config:
            log.error("No keyid or keylabel defined.")
            raise HSMException("No keyid or keylabel defined.")
        self.keyid = self.config.get("keyid")
        self.keylabel = self.config.get("keylabel")

        if "password" in self.config:
            self.password = self.config.get("password")
        else:
            log.error("No password specified.")
            self.password = getpass()
        # Now we have our password
        self.is_ready = False

        timeout = self.config.get("timeout") or DEFAULT_TIMEOUT
        lock_dir = self.config.get("lock_dir") or DEFAULT_LOCK_DIR

        log.debug("Starting Lock")
        with hsm_lock(timeout=timeout, lock_dir=lock_dir):
            log.info("Initializing PKCS11")
            self.pkcs11 = PyKCS11.PyKCS11Lib()
            self.pkcs11.load(self.config.get("module"))
            self.pkcs11.lib.C_Initialize()
            log.debug("PKCS11 initialized")

            slotlist = self.pkcs11.getSlotList()
            log.debug("Found the slots: {0!s}".format(slotlist))
            if not len(slotlist):
                raise HSMException("No HSM connected. No slots found.")

            if self.slot == -1:
                if len(slotlist) == 1:
                    # Use the first and only slot
                    self.slot = slotlist[0]
                elif len(slotlist) > 1:
                    for slot in slotlist:
                        # Find the slot via the slotname
                        slotinfo = self.pkcs11.getSlotInfo(slot)
                        log.debug("Found slot '{}'".format(slotinfo.slotDescription))
                        if slotinfo.slotDescription.startswith(self.slotname):
                            self.slot = slot
                            break
            log.info("Using slot {0!s}".format(self.slot))

            if self.slot not in slotlist:
                raise HSMException("Slot {0:d} ({1:s}) not present".format(self.slot, self.slotname))

            slotinfo = self.pkcs11.getSlotInfo(self.slot)
            log.info("Setting up slot {0!s}: '{1!s}'".format(self.slot, slotinfo.slotDescription))

            self.session = self.pkcs11.openSession(slot=self.slot)
            log.info("Logging on to '{}'".format(slotinfo.slotDescription))
            try:
                self.session.login(self.password)
            except PyKCS11.PyKCS11Error as e:
                if str(e).startswith("CKR_USER_ALREADY_LOGGED_IN"):
                    log.info("Timing issues. We need to relogin the user.")
                    # in case the user is already logged in
                    self.session.logout()
                    self.session.login(self.password)
                elif str(e).startswith("CKR_PIN_INCORRECT"):
                    log.error("A wrong HSM Password is configured. Please check your configuration in pi.cfg")
                    # We reset the password, to avoid future PIN Locking!
                    # I think this does not work between processes!
                    self.password = None
                    raise e
                else:
                    raise e
            log.info("Logged into slot {0!s}".format(self.slot))

            if "encfile" in self.config:
                self._decrypt_file(self.config.get("encfile"))
            log.info("Successfully setup the security module.")
            self.is_ready = True
            # We need this for the base class
            self.crypted = True
            if logout:
                self.session.logout()
                self.session.closeSession()

    def _add_template(self, template):
        if self.keyid:
            template.append((PyKCS11.CKA_ID, (self.keyid,)))
        elif self.keylabel:
            template.append((PyKCS11.CKA_LABEL, self.keylabel))
        return template

    def _get_private_key(self):
        """
        Returns the handle to the private key for decryption.
        """
        log.debug("Getting private key handles")
        objs = self.session.findObjects(self._add_template([(PyKCS11.CKA_CLASS, PyKCS11.CKO_PRIVATE_KEY)]))
        log.debug("Found {0!s} private keys.".format(len(objs)))
        return objs[0]

    def _encrypt_file(self, infile, outfile):
        """
        This is only used to create the encrypted file, that holds the encryption key.
        :param infile: The encryption key in plain text
        :param outfile: The encrypted encryption key
        """
        with open(infile, "rb") as f:
            enckey = f.read()
        objs = self.session.findObjects(self._add_template([(PyKCS11.CKA_CLASS, PyKCS11.CKO_PUBLIC_KEY)]))
        log.debug("Found {0!s} public keys.".format(len(objs)))
        for obj in objs:
            log.debug("========================================================")
            log.debug("Found object {0!s}".format(obj))
        pubkey = objs[0]
        m = PyKCS11.Mechanism(MECHANISM)
        r = self.session.encrypt(pubkey, enckey, m)
        with open(outfile, "wb") as f:
            f.write(bytearray(r))

    def _listkeys(self, keytype="public"):
        if keytype == "public":
            objs = self.session.findObjects([(PyKCS11.CKA_CLASS, PyKCS11.CKO_PUBLIC_KEY)])
        else:
            objs = self.session.findObjects([(PyKCS11.CKA_CLASS, PyKCS11.CKO_PRIVATE_KEY)])
        log.debug("Found {0!s} keys.".format(len(objs)))
        for obj in objs:
            log.debug("========================================================")
            log.debug("Found object {0!s}".format(obj))

    def _decrypt_file(self, filename):
        log.info("Reading encrypted key file")
        f = open(filename, "rb")
        filecontents = f.read()
        f.close()
        log.debug("Decrypting encryption keys")
        privkey = self._get_private_key()
        log.debug("Defining mechanism")
        m = PyKCS11.Mechanism(MECHANISM)
        log.debug("Calling session.decrypt with private key")
        r = self.session.decrypt(privkey, filecontents, m)
        log.debug("Keys decrypted")
        r = int_list_to_bytestring(r)
        for key_id in [0, 1, 2]:
            self.secrets[key_id] = r[key_id * 32: (key_id + 1) * 32]
        log.info("Successfully loaded encryption keys into process.")

    def _get_secret(self, slot_id=0, password=None):
        """
        This replaces the _get_secret method from the base class to return
        one of the three encryption keys.

        :param slot_id: The id of the encryption key
        :param password: n/a
        :return:
        """
        key = self.secrets.get(slot_id)
        return key

    def setup_module(self, params):
        """

        :param params:
        :return:
        """
        log.warning("The method 'setupmodule' is not implemented and can not be used.")


if __name__ == "__main__":  # pragma: no cover
    logging.basicConfig()
    log.setLevel(logging.INFO)
    log.setLevel(logging.DEBUG)

    try:
        opts, args = getopt.getopt(sys.argv[1:], "hm:s:k:i:o:e:l",
                                   ["help", "module=", "slotname=", "keyid=",
                                    "infile=", "outfile=",
                                    "encfile=", "listprivate", "listpublic", "keylabel="])
    except getopt.GetoptError as e:
        print(str(e))
        sys.exit(1)

    infile = outfile = None
    config = {}
    listkeys = False

    for o, a in opts:
        if o in ("-m", "--module"):
            config["module"] = a
        elif o in ("-s", "--slotname"):
            config["slotname"] = a
        elif o in ("-k", "--keyid"):
            config["keyid"] = int(a)
        elif o in ("--keylabel"):
            config["keylabel"] = a
        elif o in ("-i", "--infile"):
            infile = a
        elif o in ("-o", "--outfile"):
            outfile = a
        elif o in ("-e", "--encfile"):
            config["encfile"] = a
        elif o in ("-l", "--listpublic"):
            listkeys = "public"
        elif o in ("--listprivate"):
            listkeys = "private"

    p = EncryptKeyHardwareSecurityModule(config, logout=False)

    if infile and outfile:
        p._encrypt_file(infile, outfile)

    if listkeys:
        p._listkeys(listkeys)

    if "encfile" in config:
        # password
        password = "topSekr3t" * 16
        crypted = p.encrypt_password(password)
        text = p.decrypt_password(crypted)
        assert(text == password)  # nosec B101 # This is actually a test
        log.info("password encrypt/decrypt test successful")

        # pin
        password = "topSekr3t"  # nosec B105 # used for testing
        crypted = p.encrypt_pin(password)
        text = p.decrypt_pin(crypted)
        assert (text == password)  # nosec B101 # This is actually a test
        log.info("pin encrypt/decrypt test successful")

        # random
        tmp_iv = p.random(16)
        plain = p.random(128)
        log.info("random test successful")

        # generic encrypt / decrypt
        cipher = p.encrypt(plain, tmp_iv)
        assert (plain != cipher)  # nosec B101 # This is actually a test
        text = p.decrypt(cipher, tmp_iv)
        assert (text == plain)  # nosec B101 # This is actually a test
        log.info("generic encrypt/decrypt test successful")



# (c) NetKnights GmbH 2024,  https://netknights.it
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-FileCopyrightText: 2024 Henrik Falk <henrik.falk@netknights.it>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
"""This module is intended to ensure the application can be monitored effectively
and provide insights into whether the application and its dependencies are operational.
A primary intention was to create a way for Kubernetes to check on containers.

Endpoints:
    - `/healthz` : Combined health check for liveness and readiness.
    - `/healthz/startupz` : Startup check to confirm if the app has started.
    - `/healthz/livez` : Liveness check to verify if the app is running.
    - `/healthz/readyz` : Readiness check to confirm if the app is ready to serve requests.
    - `/healthz/resolversz` : Resolver check to test the connection to all LDAP and SQL resolvers.

The corresponding code is tested in tests/test_api_healthcheck.py.
"""
from flask import Blueprint, current_app

from privacyidea.api.lib.utils import send_result
from privacyidea.lib.crypto import get_hsm
from privacyidea.lib.resolver import get_resolver_list, get_resolver_class
import logging

log = logging.getLogger(__name__)

healthz_blueprint = Blueprint('healthz_blueprint', __name__)


@healthz_blueprint.route('/', methods=['GET'])
def healthz():
    """
    Combined health check endpoint for liveness and readiness.

    The health check verifies the liveness of the app, ensuring it is running,
    and the readiness, ensuring the app and its dependencies are ready to serve
    requests.

    :return: A tuple containing the readiness status and HTTP response code.
    :rtype: tuple
    """
    livez_status, livez_code = livez()
    if livez_code != 200:
        return livez_status, livez_code

    readyz_status, readyz_code = readyz()
    return readyz_status, readyz_code


@healthz_blueprint.route('/startupz', methods=['GET'])
def startupz():
    """
    Startup check endpoint that indicates if the app has started.

    This endpoint returns a status of "started" if the app is initialized and
    running, otherwise it returns "not started" with an HTTP status code of 503.

    :return: A tuple containing the startup status and HTTP response code.
    :rtype: tuple
    """
    if current_app.config.get('APP_READY'):
        return send_result({"status": "started"}), 200
    else:
        return send_result({"status": "not started"}), 503


@healthz_blueprint.route('/livez', methods=['GET'])
def livez():
    """
    Liveness check endpoint that indicates if the app is running.

    This endpoint returns an HTTP status 200 and a JSON object indicating that
    the application is live and running.

    :return: A tuple containing the liveness status and HTTP response code.
    :rtype: tuple
    """
    return send_result({"status": "OK"}), 200


@healthz_blueprint.route('/readyz', methods=['GET'])
def readyz():
    """
    Readiness check endpoint that indicates if the app is ready to serve requests.

    The endpoint checks the readiness of the app, ensuring that the app has started
    and the HSM (Hardware Security Module) is in a ready state. If any condition is
    not met, a 503 status is returned with appropriate information.

    :return: A tuple containing the readiness status and HTTP response code.
    :rtype: tuple
    """
    if not current_app.config.get('APP_READY'):
        return send_result({"status": "not ready",
                            "hsm": "unchecked"}), 503
    elif not get_hsm().is_ready:
        return send_result({"status": "not ready",
                            "hsm": "fail"}), 503
    else:
        return send_result({"status": "ready",
                            "hsm": "OK"}), 200


@healthz_blueprint.route('/resolversz', methods=['GET'])
def resolversz():
    """
    Resolver check endpoint that tests the connection to all resolvers types defined in resolver_types.
    For now LDAP and SQL resolvers are tested.

    The endpoint returns a JSON object with the status of each resolver (either "OK"
    or "fail"). It attempts to establish a connection to each resolver, and if any
    exception occurs during the check, a 503 status is returned.

    :return: A tuple containing the resolver status and HTTP response code.
    :rtype: tuple
    """
    result = {}
    resolver_types = ["ldapresolver", "sqlresolver"]
    total_status = "OK"

    try:
        for resolver_type in resolver_types:
            result[resolver_type] = {}
            resolvers_list = get_resolver_list(filter_resolver_type=resolver_type)
            for resolver_name, resolver_data in resolvers_list.items():
                if resolver_data:
                    resolver_class = get_resolver_class(resolver_type)
                    success, _ = resolver_class.testconnection(resolver_data.get("data"))
                    if not success:
                        total_status = "fail"
                    result[resolver_type][resolver_name] = "OK" if success else "fail"
                else:
                    result[resolver_type][resolver_name] = "fail"

        result["status"] = total_status
        return send_result(result), 200
    except Exception as e:
        log.debug(f"Exception in /resolversz endpoint: {e}")
        return send_result({"status": "error"}), 503



# SPDX-FileCopyrightText: (C) 2023 Jona-Samuel Höhmann <jona-samuel.hoehmann@netknights.it>
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Info: https://privacyidea.org
#
# This code is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation, either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program. If not, see <http://www.gnu.org/licenses/>.

import click
from privacyidea.cli import create_silent_app, NoPluginsFlaskGroup, get_version
from privacyidea.cli.pitokenjanitor.utils.findtokens import findtokens
from privacyidea.cli.pitokenjanitor.utils.importtokens import importtokens_cli
from privacyidea.cli.pitokenjanitor.utils.updatetokens import updatetokens

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(cls=NoPluginsFlaskGroup, create_app=create_silent_app, context_settings=CONTEXT_SETTINGS,
             add_default_commands=False, add_version_option=False,
             epilog='Check out our docs at https://privacyidea.readthedocs.io/ for more details')
@click.option(
    "--version",
    help="Show the version information",
    expose_value=False,
    callback=get_version,
    is_flag=True,
    is_eager=True,
)
def cli():
    """
\b
             _                    _______  _______
   ___  ____(_)  _____ _______ __/  _/ _ \\/ __/ _ |
  / _ \\/ __/ / |/ / _ `/ __/ // // // // / _// __ |
 / .__/_/ /_/|___/\\_,_/\\__/\\_, /___/____/___/_/ |_|
/_/                       /___/

   Management script for tokens of privacyIDEA."""


cli.add_command(findtokens)
cli.add_command(importtokens_cli)
cli.add_command(updatetokens)

if __name__ == '__main__':
    cli()
